/*	FILEMAN.CPP - File manager
*/

#include "Common.hpp"
#include <io.h>
#include <sys\stat.h>
#include <direct.h>
#include <i86.h>
#include <ASSERT.H>
#include "GrEngine.hpp"
#include "CPU.hpp"
#include "FileMan.hpp"
#include "GUI.hpp"
#include "SPC700.hpp"

char romdir [80] = "..", sramdir [80] = "..", instdir [80] = "", sshotdir [80] = "";
char exedir [80];

byte gameres = 1, guires = 4;
byte addsubmethod = 0;
byte soundkhz = 8, soundchannels = 0xFF;
boolean stereosound = false;
byte pcontrol[4] = { 0, 1, 3, 3 };
boolean autobacktrack = true, disallowinst = false;
word backtracktime = 30, backtracknum = 3;
struct windowsettings ws [NUMWINDOWS];

extern struct filemap fm[FTAGS] = {
	"CheatDesc", FCHEAT, FSTRINGALLOC, offsetof (cheatpatchset, desc),     // 0
	"AutoRegOn", FROMSETTING, FBYTE, offsetof (gamesettings, autoreg),
	"FrameSkip", FROMSETTING, FBYTE, offsetof (gamesettings, frameskip),
	"Regulate%", FROMSETTING, FBYTE, offsetof (gamesettings, regpercent),
	"GrSpeedup", FROMSETTING, FBYTE, offsetof (gamesettings, grspeedup),
	"65c816%",   FROMSETTING, FBYTE, offsetof (gamesettings, cpupercent),  // 5
	"SPC700%",   FROMSETTING, FBYTE, offsetof (gamesettings, spcpercent),
	"SoundChs",  FROMSETTING, FWORD, offsetof (gamesettings, soundch),
	"BGs",       FROMSETTING, FBYTE, offsetof (gamesettings, bgs),
	"ForceROM",  FROMSETTING, FBYTE, offsetof (gamesettings, forcerom),
	"ForceCtry", FROMSETTING, FBYTE, offsetof (gamesettings, forcecountry), // 10
	"Interl",    FROMSETTING, FBYTE, offsetof (gamesettings, interleaved),
	"EnableSPC", FROMSETTING, FBYTE, offsetof (gamesettings, emulatespc),
	"SPCSkipM",  FROMSETTING, FBYTE, offsetof (gamesettings, spcskipmethod),
	"OpenWins",  FROMSETTING, FWORD, offsetof (gamesettings, openwindows),
	"NoFastROM", FROMSETTING, FBYTE, offsetof (gamesettings, ignorefastrom),
	"GameRes",   FGLOBAL, FBYTE, (dword) &gameres,        // 15
	"GUIRes",    FGLOBAL, FBYTE, (dword) &guires,
	"AddSubM",   FGLOBAL, FBYTE, (dword) &addsubmethod,
	"SoundKHz",  FGLOBAL, FBYTE, (dword) &soundkhz,
	"StereoSnd", FGLOBAL, FBYTE, (dword) &stereosound,
	"P1Control", FGLOBAL, FBYTE, (dword) &pcontrol[0],      // 20
	"P2Control", FGLOBAL, FBYTE, (dword) &pcontrol[1],
	"P3Control", FGLOBAL, FBYTE, (dword) &pcontrol[2],
	"P4Control", FGLOBAL, FBYTE, (dword) &pcontrol[3],
	"DisableIS", FGLOBAL, FBYTE, (dword) &disallowinst,
	"WaitTrace", FGLOBAL, FBYTE, (dword) &waitretrace,    // 25
	"LeadingEd", FGLOBAL, FBYTE, (dword) &leadingedge,
	"ROMDir",    FGLOBAL, FSTRING, (dword) romdir,
	"SRAMDir",   FGLOBAL, FSTRING, (dword) sramdir,
	"InstDir",   FGLOBAL, FSTRING, (dword) instdir,
	"SShotDir",  FGLOBAL, FSTRING, (dword) sshotdir,      // 30
	"BackGrnd",  FGLOBAL, FBYTE, (dword) &background,
	"ESCQuits",  FGLOBAL, FBYTE, (dword) &escquits,
	"WindowSet", FGLOBAL, FBYTE, (dword) &set.window,
	"TextSet",   FGLOBAL, FBYTE, (dword) &set.text,
	"CtrlSet",   FGLOBAL, FBYTE, (dword) &set.control,     // 35
	"ChoiceSet", FGLOBAL, FBYTE, (dword) &set.choice,
	"ButtonSet", FGLOBAL, FBYTE, (dword) &set.button,
	"BtTextSet", FGLOBAL, FBYTE, (dword) &set.buttontext,
	"TitleSet",  FGLOBAL, FBYTE, (dword) &set.title,
	"TtTextSet", FGLOBAL, FBYTE, (dword) &set.titletext,   // 40
	"FocusSet",  FGLOBAL, FBYTE, (dword) &set.focus,
	"FlashSet",  FGLOBAL, FBYTE, (dword) &set.flash,
	"JoyUp",     FJOYPADSETTING, FBYTE, offsetof (struct snesjoypad, up),
	"JoyDown",   FJOYPADSETTING, FBYTE, offsetof (struct snesjoypad, down),
	"JoyLeft",   FJOYPADSETTING, FBYTE, offsetof (struct snesjoypad, left),   // 45
	"JoyRight",  FJOYPADSETTING, FBYTE, offsetof (struct snesjoypad, right),
	"JoyX",      FJOYPADSETTING, FBYTE, offsetof (struct snesjoypad, x),
	"JoyY",      FJOYPADSETTING, FBYTE, offsetof (struct snesjoypad, y),
	"JoyA",      FJOYPADSETTING, FBYTE, offsetof (struct snesjoypad, a),
	"JoyB",      FJOYPADSETTING, FBYTE, offsetof (struct snesjoypad, b),      // 50
	"JoyL",      FJOYPADSETTING, FBYTE, offsetof (struct snesjoypad, l),
	"JoyR",      FJOYPADSETTING, FBYTE, offsetof (struct snesjoypad, r),
	"JoySelect", FJOYPADSETTING, FBYTE, offsetof (struct snesjoypad, select),
	"JoyStart",  FJOYPADSETTING, FBYTE, offsetof (struct snesjoypad, start),
	"WinPosX",   FWINDOWPOS, FWORD, offsetof (struct windowsettings, x),      // 55
	"WinPosY",   FWINDOWPOS, FWORD, offsetof (struct windowsettings, y),
	"WinZOrder", FWINDOWPOS, FBYTE, offsetof (struct windowsettings, zorder),
	"JoyMinX",   FGLOBAL, FWORD, (dword) &joyminx,
	"JoyMaxX",   FGLOBAL, FWORD, (dword) &joymaxx,
	"JoyMinY",   FGLOBAL, FWORD, (dword) &joyminy,         // 60
	"JoyMaxY",   FGLOBAL, FWORD, (dword) &joymaxy,
	"JoyMidX",   FGLOBAL, FWORD, (dword) &joymidx,
	"JoyMidY",   FGLOBAL, FWORD, (dword) &joymidy,
	"JoyThresh", FGLOBAL, FWORD, (dword) &joythresh,
	"BackNum",   FGLOBAL, FWORD, (dword) &backtracknum,    // 65
	"FPUCopy",   FGLOBAL, FBYTE, (dword) &fpucopy,        
	"Windowing", FROMSETTING, FBYTE, offsetof (gamesettings, windowing),
	"IndirHDMA", FROMSETTING, FBYTE, offsetof (gamesettings, indirecthdma),
	"AutoBackT", FGLOBAL, FBYTE, (dword) &autobacktrack,
	"BackTTime", FGLOBAL, FWORD, (dword) &backtracktime,
	"AskB4Exit", FGLOBAL, FBYTE, (dword) &confirmexit,
	"ShowFPS",   FGLOBAL, FBYTE, (dword) &showfps,
};

	// Commands:
	//
	// New ROM "filename.ext"
	// New Joypad n
	// New Window n
	// New Cheat bbaaaa=vv+...
	//
	// ; Statements starting with a semicolon are ignored.

struct gamesettings gsdefault = {
	"[Defaults]", // char filename [13];
	true,     // boolean autoreg;
	2, 100,   // byte frameskip, regpercent;
	false,    // boolean grspeedup;
	100, 100, // byte cpupercent, spcpercent;
	0xFF,     // word soundch;
	31,       // byte bgs;
	0, 0,     // byte forcerom, forcecountry;
	false,    // boolean interleaved;
	false,    // boolean emulatespc;
	0,        // byte spcskipmethod;
	1,        // word openwindows;
	true,     // boolean ignorefastrom
	true, true// boolean windowing, indirecthdma
};

struct gamesettings *gs [MAXGAMEDATABASE] = {
	&gsdefault // First entry (gs[0]) is the defaults.
};
struct gamesettings *curgs;
int curgsnum = 0;
int numgamesettings = 0;
int gsentries = 1;
boolean romloaded = false, romrunning = false;

void initfilemanager (char *argv0)
{
	char temp[80];
	
	curgs = gs[curgsnum];
	strcpy (temp, argv0);
	stripfilename (temp);
	_fullpath (exedir, temp, 80);
	debug0(exedir);
}

void uninitfilemanager (void)
{
	int x, y;

	for (x = 1; x < MAXGAMEDATABASE; x++) {
		if (gs[x] == NULL) continue;
		for (y = 0; y < gs[x]->patches; y++) {
			free (gs[x]->cheat[y]->desc);
			free (gs[x]->cheat[y]->patch);
			free (gs[x]->cheat[y]);
		}
		free (gs[x]);
	}
}

boolean exists (char *s)
{
	if (access (s, 0) == 0)
		return true;
	else
		return false;
}

boolean isadir (char *s)
{
	static struct _stat buf;

	// first checks whether the path is valid...
    if (_stat (s, &buf) == -1) {
		// no, so it might be an incomplete path.  Decide based on the existance of an extention.
		if (findlastch (s + findlastch(s, SLASH) + 1, '.') == -1) {
			return true;
		} else {
			return false;
		}
	} else if (S_ISDIR(buf.st_mode))
		return true;
	return false;
}

void delstr (char *s, int p, int n)
{
	for (s += p; *s != '\0'; s++) {
		*s = s[n];
		if (*s == '\0') break;
	}
}

int findlastch (char *s, char ch)
{
	int x;
	for (x = strlen(s)-1; x >= 0; x--) {
		if (s[x] == ch) {
			return x;
		}
	}
	return x;
}

int findfirstch (char *s, char ch)
{
	int x;
	for (x = 0; s[x] != '\0'; x++) {
		if (s[x] == ch) {
			return x;
		}
	}
	return -1;
}

int countchar (char *s, char ch)
{
	int n = 0;
	
	while (*s != '\0') {
		if (*s == ch) n++;
		s++;
	}
	return n;
}

void trimspace (char *s)
{
	int y, n;
	
	for (y = 0; s[y] == ' ' || s[y] == 9; y++)
		;
	delstr (s, 0, y);
	for (y = strlen(s)-1, n = 0; (s[y] == ' ' || s[y] == 9) && y >= 0; y--, n++)
		;
	delstr (s, y+1, n);
}

void trimfilespec (char *s)
{
	trimspace (s);
	if (countchar (s, SLASH) > 1 && s [strlen(s) - 1] == SLASH) {
		s [strlen(s) - 1] = '\0';
	}
}

void strippathname (char *name)
{
	int x, p;

	if (isadir (name)) {
		name[0] = '\0';
		return;
	}
	p = findlastch (name, SLASH);
	x = findlastch (name, ':');
	if (x > p) p = x;
	delstr (name, 0, p + 1);
}

void stripfilename (char *name)
{
	int x;
	if (!isadir (name)) {
		x = findlastch (name, SLASH);
		if (name [x - 1] == ':' || x == 0) {
			x++; // Don't delete backslash if file is in root.
		}
		delstr (name, x, strlen (name) - x);
	} else {
		x = strlen(name);
		if (name[x-1] == SLASH)
			name[x-1] = '\0';
	}
}

void getabspath (char *path)
{	// absolute path reletive to executable directory
	char tempcwd [80], temppath [80];
	unsigned dummy;
	
	getcwd (tempcwd, 80);
	_dos_setdrive (exedir [0]-'A'+1, &dummy);
	chdir (exedir);
	_fullpath (temppath, path, 80);
	strcpy (path, temppath);
	_dos_setdrive (tempcwd [0]-'A'+1, &dummy);
	chdir (tempcwd);
}

void combinepathfile (char *comb, char *path, char *fn)
{	// This function doesn't combine two paths... it just appends fn to path
	int x = strlen (path);
	
	strcpy (comb, path);
	if (comb [x-1] != SLASH) {
		comb [x] = SLASH;
		x++;
	}
	strcpy (comb+x, fn);
}

void combinepaths (char *comb, char *path, char *offs)
{
	char tempcwd [80], temppath [80];
	unsigned dummy, originaldrive;
	
	_dos_getdrive (&originaldrive); // Save current drive
	path[0] = toupper(path[0]);
	offs[0] = toupper(offs[0]);
	if (offs[1] == ':')
		_dos_setdrive (offs[0]-'A'+1, &dummy);
	else if (path [1] == ':')
		_dos_setdrive (path[0]-'A'+1, &dummy);
	getcwd (tempcwd, 80); // save current directory

	_fullpath (temppath, path, 80);
	_dos_getdrive (&dummy);
	if (dummy == temppath[0]-'A'+1)
		chdir (path);
	chdir (offs);
	getcwd (comb, 80);

	_dos_setdrive (tempcwd [0]-'A'+1, &dummy);
	chdir (tempcwd); // restore current directory
	_dos_setdrive (originaldrive, &dummy); // Restore current drive
}

#define ntohex(n) ((n) > 9 ? (n)+'A'-10 : (n)+'0')
#define ishexd(c) (((c) >= '0' && (c) <= '9') || ((c) >= 'A' && (c) <= 'Z'))
#define hexton(c) ((c) >= 'A' ? (c)-'A'+10 : (c)-'0')
	// ONLY WORKS ON UPPERCASE HEX!

int verifyspecformat (char *patchspec)
{	// Removes spaces...btw...and converts to uppercase.
	// False=bad format, other=number of patches.
	char *o_patchspec = patchspec;
	int i, n;
	
	strupr (patchspec);
	debug0 ("Checking %s",patchspec);
	while (*patchspec)
		if (*patchspec == ' ')
			for (i = 0; patchspec[i]; i++)
				patchspec[i] = patchspec[i+1];
		else
			patchspec++;
	patchspec = o_patchspec;
	i = 0; n = 0;
	for (;;) {
		if (!((patchspec[i + 4] == '-')  ^ (patchspec[i + 6] == '='))) {
			debug0 ("Not game genie nor native format");
			return false;  // Must have a dash or equals, but not both
		}
		if (!ishexd(patchspec[i]) || !ishexd(patchspec[i+1]) || !ishexd(patchspec[i+2]) ||
			!ishexd(patchspec[i+3]) || (!ishexd(patchspec[i+4]) && patchspec[i+4] != '-') ||
			!ishexd(patchspec[i+5]) || (!ishexd(patchspec[i+6]) && patchspec[i+6] != '=') ||
			!ishexd(patchspec[i+7]) || !ishexd(patchspec[i+8])) {
			debug0 ("Digit problem in cheat");
			return false;  // Must all be hex digits
		}
		n++;
		if (patchspec[i + 6] == '=' && ishexd(patchspec[i+9]) && ishexd(patchspec[i+10])) {
			i += 2; // 2 More digits.
			debug0 ("Word-sized cheat");
		}
		if (strlen (patchspec) - i < 10)
			break;
		if (patchspec[i+9] != '+') {
			debug0 ("Multiple patches detected in cheat");
			return false;
		}
		i += 10;
		debug0 ("i+10");
	}
	if (strlen (patchspec) - i != 9)
		return false;
	return n;
}

int createpatch (char *patchspec, char *desc, boolean gui)
{	// Patch is created in curgs rom setting
	// Patchspec will be converted to uppercase and have spaces removed.
	// Returns -1 if the patchspec is invalid or out of memory
	// NULL description accepted.
	struct cheatpatchset *cheat;
	byte value;
	dword addr;
	int howmany, x;
	
	if (curgs->patches >= MAXPATCHSETS) {
		addmessage ("Exceeded maximum of %d cheat entries!", MAXPATCHSETS);
	}
	//debug0 ("Createpatch entered.");
	if ((howmany = verifyspecformat (patchspec)) == false)
		return -1;
	//debug0 ("Howmany: %d", howmany);
	if ((curgs->cheat[curgs->patches] = (struct cheatpatchset *)     malloc (sizeof (struct cheatpatchset))) == NULL ||\
		 (curgs->cheat[curgs->patches]->patch = (struct cheatpatch *) malloc (sizeof (struct cheatpatch) * howmany)) == NULL) {
		if (gui) addmessage ("OUT OF MEMORY!");
		else printf ("OUT OF MEMORY!");
		return -1;
	}
	if (desc != NULL)
		if ((curgs->cheat[curgs->patches]->desc = (char *) malloc (strlen (desc) + 1)) == NULL) {
			if (gui) addmessage ("OUT OF MEMORY!");
			else printf ("OUT OF MEMORY!");
			return -1;
		} else {
			strcpy (curgs->cheat[curgs->patches]->desc, desc);
		}
	else
		curgs->cheat[curgs->patches]->desc = NULL;
	//debug0 ("desc:%P curgs->cheat[curgs->patches]->desc %P", desc, curgs->cheat[curgs->patches]->desc);
	cheat = curgs->cheat[curgs->patches];
	for (x = 0; x < howmany; x++) {
		if (patchspec[4] == '-') {
			//debug0 ("%d Game Genie!", x);
			getggmeaning (patchspec, value, addr);
			//debug0 ("Value:%d addr:%X", value, addr);
			cheat->patch[x].wordsized = false;
			//debug0 ("curgs->cheat[curgs->patches]->patch[0].wordsized %d", curgs->cheat[curgs->patches]->patch[0].wordsized);
			cheat->patch[x].newval = value;
			cheat->patch[x].prevval = *SNESMEM (addr);
			patchspec += 10;
		} else {
			//debug0 ("%d Raw patch!", x);
			addr = (hexton(patchspec[0]) << 20) + (hexton(patchspec[1]) << 16) + \
			       (hexton(patchspec[2]) << 12) + (hexton(patchspec[3]) << 8) + \
			       (hexton(patchspec[4]) <<  4) +  hexton(patchspec[5]);
			if (ishexd(patchspec[9]) && ishexd(patchspec[10])) {
				//debug0 ("Word patch detected!", x);
				cheat->patch[x].newval = \
			       (hexton(patchspec[7]) << 12) + (hexton(patchspec[8]) << 8) + \
			       (hexton(patchspec[9]) <<  4) +  hexton(patchspec[10]);
				cheat->patch[x].wordsized = true;
				patchspec += 12;
			} else {
				cheat->patch[x].newval = \
			       (hexton(patchspec[7]) <<  4) +  hexton(patchspec[8]);
				cheat->patch[x].wordsized = false;
				patchspec += 10;
			}
		}
		cheat->patch[x].bank = addr >> 16;
		cheat->patch[x].addr = addr & 0xFFFF;
	}
	//		debug0 ("curgs->patches = %d curgs->cheat[curgs->patches]->patch[0].wordsized %d",curgs->patches, curgs->cheat[curgs->patches]->patch[0].wordsized);
	//debug0 ("curgs->cheat[curgs->patches]->desc %P", curgs->cheat[curgs->patches]->desc);
	curgs->patches++;
	cheat->addrs = howmany;
	cheat->enabled = false;
	return curgs->patches - 1;
}

void deletepatch (int n)
{
	int x;
	
	assert (n < curgs->patches);
	free (curgs->cheat[n]->desc);
	free (curgs->cheat[n]->patch);
	free (curgs->cheat[n]);
	curgs->patches--;
	for (x = n; x < curgs->patches; x++) {
		curgs->cheat[x] = curgs->cheat[x+1];
	}
}

static byte gghex [16] = {
	13, 15, 4, 7, 0, 9, 1, 5, 6, 11, 12, 8, 10, 2, 3, 14
};
static byte ggaddrmap [24] = {
	'i','j','k','l','q','r','s','t',
	'o','p','a','b','c','d','u','v',
	'w','x','e','f','g','h','m','n'
};

void getggmeaning (char *strcode, byte &value, dword &addr)
{	// strcode will be converted to uppercase.  It MUST be in the format xxxx-xxxx.
	int x;
	dword code, code2;

	strupr (strcode);
	code  = (gghex [hexton (strcode[0])] << 28);
	code |= (gghex [hexton (strcode[1])] << 24);
	code |= (gghex [hexton (strcode[2])] << 20);
	code |= (gghex [hexton (strcode[3])] << 16);
	code |= (gghex [hexton (strcode[5])] << 12);
	code |= (gghex [hexton (strcode[6])] << 8);
	code |= (gghex [hexton (strcode[7])] << 4);
	code |=  gghex [hexton (strcode[8])];
	debug0 ("Code:%X", code);
	code2 = 0;
	for (x = 0; x < 24; x++) {
		code2 |= ((code >> x) & 1) << ('x' - ggaddrmap [23-x]);
	}
	value = code >> 24;
	addr = code2 & 0x00FFFFFF;
}

int reversegghex (byte n)
{
	int y;
	for (y = 0; y < 16; y++)
		if (gghex[y] == n)
			return y;
	debug0 ("ERROR TRANSLATING TO GG CODE (n=%d)", n);
	return 0;
}

char *getggcode (byte value, dword addr)
{
	static char codestr [10];
	dword code, code2;
	int x, y;

	code = (value << 24) + addr;
	code2 = value << 24;
	for (x = 0; x < 24; x++) {
		code2 |= ((code >> ('x' - ggaddrmap [23-x])) & 1) << x;
	}
	y = reversegghex ((code2 >> 28) & 15); codestr [0] = ntohex (y);
	y = reversegghex ((code2 >> 24) & 15); codestr [1] = ntohex (y);
	y = reversegghex ((code2 >> 20) & 15); codestr [2] = ntohex (y);
	y = reversegghex ((code2 >> 16) & 15); codestr [3] = ntohex (y);
	y = reversegghex ((code2 >> 12) & 15); codestr [5] = ntohex (y);
	y = reversegghex ((code2 >>  8) & 15); codestr [6] = ntohex (y);
	y = reversegghex ((code2 >>  4) & 15); codestr [7] = ntohex (y);
	y = reversegghex (code2 & 15);         codestr [8] = ntohex (y);
	codestr [4] = '-';
	codestr [9] = '\0';
	return codestr;
}

char *getpatchname (int pn, boolean ggformat)
{	// Patch is returned in a STATIC char buffer.
	static char s [80];
	struct cheatpatch *c = curgs->cheat[pn]->patch;
	int y, p;

	if (pn >= curgs->patches) {
		s[0] = '\0'; return s;
	}
	if (ggformat) {
		for (y = 0, p = 0; y < curgs->cheat[pn]->addrs; y++) {
			if (c[y].wordsized) {
				strcpy (s + p, getggcode (c[y].newval >> 8, (c[y].bank << 16) + c[y].addr + 1));
				s [p + 9] = '+';
				p += 10;
			}
			strcpy (s + p, getggcode (c[y].newval, (c[y].bank << 16) + c[y].addr));
			s [p + 9] = '+';
			p += 10;
		}
		s[p-1] = '\0';
	} else {
		for (y = 0, p = 0; y < curgs->cheat[pn]->addrs; y++) {
			debug0 ("Starting #%d",y);
			s[p++] = ntohex (c[y].bank >> 4);
			s[p++] = ntohex ((c[y].bank) & 15);
			s[p++] = ntohex (c[y].addr >> 12);
			s[p++] = ntohex ((c[y].addr >> 8) & 15);
			s[p++] = ntohex ((c[y].addr >> 4) & 15);
			s[p++] = ntohex (c[y].addr & 15);
			s[p++] = '=';
			if (c[y].wordsized) {
				s[p++] = ntohex (c[y].newval >> 12);
				s[p++] = ntohex ((c[y].newval >> 8) & 15);
			}
			s[p++] = ntohex ((c[y].newval >> 4) & 15);
			s[p++] = ntohex (c[y].newval & 15);
			s[p] = '+';
			if (y != curgs->cheat[pn]->addrs - 1)
				p++;
		}
		s[p] = '\0';
	}
	debug0 ("Done!");
	return s;
}

int findgsnum (char *filename)
{
	int y, sp, cp;

	sp = findlastch (filename, SLASH) + 1;
	cp = findlastch (filename, ':') + 1;
	if (cp > sp) sp = cp;
	for (y = 0; y < gsentries; y++) {
		if (stricmp (filename + sp, gs[y]->filename) == 0) {
			return y;
		}
	}
	return 0; // Not found/default
}

void createromentry (char *filename)
{
	int y, sp, cp;

	sp = findlastch (filename, SLASH) + 1;
	cp = findlastch (filename, ':') + 1;
	if (cp > sp) sp = cp;
	debug0 ("createromentry %s", filename + sp);
	y = findgsnum (filename);
	if (y != 0 || (y == 0 && stricmp (filename + sp, "[Defaults]") == 0)) {
		curgsnum = y;
		curgs = gs[y];
		debug0 ("found existing entry");
		return;
	} else
		y = gsentries;
	gsentries++;
	debug0 ("making new entry %d (y=%d)", gsentries, y);
	gs [y] = (struct gamesettings *) malloc (sizeof (struct gamesettings));
	curgsnum = y;
	curgs = gs[y];
	memcpy (gs[y], gs[0], sizeof (struct gamesettings)); // Load default values
	strcpy (gs[y]->filename, filename + sp);
}

static int lline;

void loadline (FILE *fp, char *l)
{
	int x;
	boolean inquote;
	
	do {
		inquote = false;
		for (x = 0; x < 100; x++) {
			l [x] = fgetc (fp);
			if (l [x] == '"')
				inquote = !inquote;
			if (l [x] == ';' && !inquote)
				l [x] = '\0';
			if (l [x] == '\n' || feof (fp)) {
				l [x] = '\0';
				lline++;
				break;
			}
		}
		trimspace (l);
	} while (l [0] == '\0' && !feof (fp));
	//debug0 ("line loaded: %s len: %d", l, strlen (l));
}

void loadsettings (char *filename)
{
	FILE *fp;
	byte *addr = NULL;
	char full[80], l[100], utag[10];
	int x, t, ep, prevgsnum = curgsnum;
	int curjoy = 0, curwin = 0, curcheat = 0;
	boolean inquote;

	strcpy (full, filename);
	getabspath (full);
	printf ("Loading configuration file: %s\n", full);
	debug0 ("Loading configuration file: %s (saved curgsnum: %d)", full, curgsnum);
	lline = 0;
	for (x = 1; x < NUMWINDOWS; x++)
		win[x]->zorder = -1; // Make sure windows are closed by default...
	storewinsettings (); // get defaults in case config file doesn't load it
	if ((fp = fopen (full, "rt")) == NULL) {
		debug0 ("Error opening config file (%s) to load!", full);
		printf ("Error opening config file (%s) to load!\nOperation aborted.", full);
		sleep (1);
		return;
	}
	printf ("Opened successfully\n");
	do {
		loadline (fp, l);
		for (inquote = 0, x = 0; l [x] != '\0'; x++) {
			if (l [x] == '"')
				inquote = !inquote;
			if (!inquote)
				l [x] = toupper (l [x]);
		}
		if (l[0] == '\0')
			continue;
		if (l[0] == 'N' && l[1] == 'E' && l[2] == 'W' && l[3] == ' ') {
			delstr (l, 0, 4);
			trimspace (l);
			ep = findfirstch (l, ' ');
			if (ep < 0) ep = 0;
			l[ep] = '\0';
			ep++;
			trimspace (l + ep);
			if (strcmp (l, "ROM") == 0) {
				if (l[ep] != '"' || l[ep + strlen(l + ep) - 1] != '"') {
					printf ("Warning! Line %d: Missing or misplaced quote mark(s) - Ignoring\n", lline);
					printf ("   Note! ROM file spec was: %s\n", l + ep);
				} else {
					delstr (l + ep, 0, 1);
					delstr (l + ep, strlen(l + ep) - 1, 1);
					createromentry (l + ep);
				}
			} else if (strcmp (l, "JOYPAD") == 0) {
				curjoy = atoi (l + ep);
				//debug0 ("New Joypad %d", curjoy);
			} else if (strcmp (l, "WINDOW") == 0) {
				curwin = atoi (l + ep);
				//debug0 ("New Window %d", curwin);
			} else if (strcmp (l, "CHEAT") == 0) {
				if (toupper(l[ep]) == 'O' && toupper(l[ep+1]) == 'N') {
					ep += 3;
					x = true;
				} else
					x = false;
				// Create new cheat code...
				curcheat = createpatch (l + ep, NULL, false);
				if (curcheat == -1) {
					printf ("Error in cheat code format!  Cheat patch not created.");
					fflush (stdout);
				} else {
					curgs->cheat[curcheat]->enabled = x;
					//debug0 ("New Cheat %d %s", curcheat, l+ep);
					//debug0 ("curcheat%d curgs->patches = %d curgs->cheat[curgs->patches-1]->patch[0].wordsized %d",curcheat,curgs->patches, curgs->cheat[curgs->patches-1]->patch[0].wordsized);
					//debug0 ("curgs->patches:%d Addrs:%d enabled:%d [0].wordsized:%d [0].bank %X [0].addr %X [0].newval %X desc%P",
					//	curgs->patches, curgs->cheat[curcheat]->addrs, curgs->cheat[curcheat]->enabled,
					//	curgs->cheat[curcheat]->patch[0].wordsized, curgs->cheat[curcheat]->patch[0].bank,
					//	curgs->cheat[curcheat]->patch[0].addr, curgs->cheat[curcheat]->patch[0].newval, curgs->cheat[curcheat]->desc);
				}
			} else {
				printf ("Warning! Line %d: Unrecognised or missing NEW spec - Ignoring\n", lline);
				printf ("   Note! NEW spec was: \"%s\"\n", l);
			}
		} else if ((ep = findfirstch (l, '=')) > 0) {
			l[ep] = '\0';
			ep++;
			//debug0 ("Var tag: %s Equals: %s", l, l+ep);
			// Search for variable tag in fm array
			for (t = 0; t < FTAGS; t++) {
				strcpy (utag, fm [t].tag);
				strupr (utag);
				if (strcmp (utag, l) == 0)
					break;
			}
			if (t >= FTAGS) {
				printf ("Warning! Line %d: Unrecognised variable - Ignoring\n", lline);
				printf ("   Note! Variable was: %s\n", l);
			} else {
				//debug0("fm [t].datatype %d name %s (%d)", fm [t].datatype, l, FSTRINGALLOC);
				switch (fm [t].datastruct) {
				case FGLOBAL:
					addr = (byte *) fm[t].offset;
					break;
				case FROMSETTING:
					addr = (((byte *) curgs) + fm[t].offset);
					break;
				case FJOYPADSETTING:
					addr = (((byte *) &joykey[curjoy]) + fm[t].offset);
					break;
				case FCHEAT:
					addr = (((byte *) curgs->cheat[curcheat]) + fm[t].offset);
					break;
				case FWINDOWPOS:
					addr = (((byte *) &ws[curwin]) + fm[t].offset);
					break;
				}
				if (fm [t].datatype == FSTRING || fm [t].datatype == FSTRINGALLOC) {
					
					if (l[ep] != '"' || l[ep + strlen(l + ep) - 1] != '"') {
						printf ("Warning! Line %d: Missing or misplaced quote mark(s) - Ignoring\n", lline);
						printf ("   Note! ROM file spec was: %s\n", l + ep);
					} else {
						//debug0 ("FSTRING %s", l+ep);
						delstr (l + ep, 0, 1);
						delstr (l + ep, strlen(l + ep) - 1, 1);
						//debug0 ("FSTRING trimmed %s", l+ep);  //stopped here..
						//debug0 ("String %s", l);
						if (fm [t].datatype == FSTRINGALLOC) {
							//debug0 ("Allocating %d bytes for %s (curcheat=%d)", strlen (l + ep) + 1, l+ep, curcheat);
							if (*(char**)addr != NULL) {
								debug0 ("Freeing %P!", *(void**)addr);
								free (*(void**)addr);
							}
							*(char **)addr = (char *) malloc (strlen (l + ep) + 1);
							//debug0 ("Copied");
							strcpy (*(char **) addr, l + ep);
							//debug0 ("Copied");
						} else
							strcpy ((char *) addr, l + ep);
					}
				} else {
					x = atoi (l + ep);
					//debug0 ("Setting variable %s to %d (fm[%d])", l, x, t);
					switch (fm [t].datatype) {
					case FBYTE:
						*addr = x;
						break;
					case FWORD:
						*(word *)addr = x;
						break;
					case FDWORD:
						*(dword *)addr = x;
						break;
					}
				}
			}
		} else {
			printf ("Warning! Line %d: Syntax error (Assignment or NEW expected) - Ignoring\n", lline);
			printf ("   Note! Line was: %s\n", l);
			debug0 ("   Warning on line %d which was: %s", lline, l);
		}
	} while (!feof (fp));
	fclose (fp);
	debug0 ("Exiting loadsettings(). Restoring curgsnum to %d", prevgsnum);
	curgsnum = prevgsnum;
	curgs = gs[curgsnum];
}

void savetag (FILE *fp, int t, int e = 0, int c = 0)
{
	byte *addr = NULL;
	
	fprintf (fp, "%s=", fm[t].tag);
	switch (fm[t].datastruct) {
	case FGLOBAL:
		addr = (byte *) fm[t].offset;
		break;
	case FROMSETTING:
		addr = (((byte *) gs[e]) + fm[t].offset);
		break;
	case FJOYPADSETTING:
		addr = (((byte *) &joykey[e]) + fm[t].offset);
		break;
	case FCHEAT:
		addr = (((byte *) gs[e]->cheat[c]) + fm[t].offset);
		break;
	case FWINDOWPOS:
		addr = (((byte *) &ws[e]) + fm[t].offset);
		break;
	}
	switch (fm[t].datatype) {
	case FBYTE:
		fprintf (fp, "%d\n", *addr);
		break;
	case FWORD:
		fprintf (fp, "%d\n", *((word *)addr));
		break;
	case FDWORD:
		fprintf (fp, "%d\n", *((dword *)addr));
		break;
	case FSTRING:
		fprintf (fp, "\"%s\"\n", addr);
		break;
	case FSTRINGALLOC:
		fprintf (fp, "\"%s\"\n", *((char **)addr));
		break;
	}
}

void savesettings (char *filename)
{
	FILE *fp;
	char full[80];
	int x, y, z, prevgsnum = curgsnum;

	debug0 ("About to save config file");
	strcpy (full, filename);
	getabspath (full);
	printf ("Saving configuration file: %s\n", full);
	if ((fp = fopen (full, "wt")) == NULL) {
		debug0 ("Error opening config file (%s) to save!", full);
		printf ("Error opening config file (%s) to save! Operation aborted.", full);
		return;
	}
	debug0 ("Open config file to save");
	fprintf (fp, ";\n; %s configuration file - automatically generated by %s.\n;\n", PROGRAMNAME, PROGRAMNAME);
	fprintf (fp, "\n; Misc. Global definitions\n");
	for (y = 0; y < FTAGS; y++) {
		fflush (fp);
		if (fm[y].datastruct == FGLOBAL)
			savetag (fp, y);
	}
	fprintf (fp, "\n; ROM settings: If the ROM database gets too large (>250 entries),\n; %s will begin to malfunction.  If this happens, delete the option sets for a few\n; of your ROMs that you never use. (and don't tell me you play more than 250 games.)\n", PROGRAMNAME);
	for (x = 0; x < gsentries; x++) {
		curgs = gs[x];
		curgsnum = x;
		fprintf (fp, "\nNew ROM \"%s\"\n", gs[x]->filename);
		fflush (fp);
		for (y = 0; y < FTAGS; y++) {
			if (fm[y].datastruct == FROMSETTING)
				savetag (fp, y, x);
			if (fm[y].datastruct == FCHEAT) {
				for (z = 0; z < gs[x]->patches; z++) {
					fprintf (fp, "New Cheat%s %s\n", curgs->cheat[z]->enabled ? " On" : "", getpatchname (z, false));
					savetag (fp, y, x, z); 
				}
			}
			fflush (fp);
		}
	}
	fprintf (fp, "\n; Joypad settings: The scancodes for each of the six Joypad configurations.\n");
	for (x = 0; x < 6; x++) {
		fprintf (fp, "\nNew Joypad %d\n", x);
		for (y = 0; y < FTAGS; y++) {
			if (fm[y].datastruct == FJOYPADSETTING)
				savetag (fp, y, x);
			fflush (fp);
		}
	}
	fprintf (fp, "\n; Window positions for the GUI windows.\n");
	for (x = 0; x < NUMWINDOWS; x++) {
		fprintf (fp, "\nNew Window %d\n", x);
		for (y = 0; y < FTAGS; y++) {
			if (fm[y].datastruct == FWINDOWPOS)
				savetag (fp, y, x);
			fflush (fp);
		}
	}
	fprintf (fp, "\n; End of file - finally!\n");
	fclose (fp);
	curgsnum = prevgsnum;
	curgs = gs[curgsnum];
}

void storewinsettings ()
{
	int x, y;

	//debug0 ("Storing win settings...");
	for (y = 0; y < NUMWINDOWS; y++) {
		ws [y].zorder = win[y]->zorder;
		ws [y].x = win[y]->x;
		ws [y].y = win[y]->y;
	}
	curgs->openwindows = 0;
	for (x = 1; x < NUMWINDOWS; x++) {
		if (win[x]->zorder >= 0)
			curgs->openwindows |= (1 << x);
	}
}

void applywinsettings ()
{
	int y;
	signed char x;

	for (y = 0; y < NUMWINDOWS; y++) {
		win[y]->x = ws [y].x;
		win[y]->y = ws [y].y;
	}
	for (x = 127; x > -100; x--) {
		for (y = 1; y < NUMWINDOWS; y++) {
			if (ws [y].zorder == x) {
				if ((curgs->openwindows & (1 << y)) && (y != 7 || curgsnum == 0)) {
					showwindow (y);
				} else {
					killwindow (y);
				}
			}
		}
	}
}

void loadromheaderdata (char *fn, struct headerdata *h, int forcerom)
{
	word c[2];
	byte b;
	FILE *fp;

	debug0 ("  Loadromheaderdata entered.");
	h->hadd = 512; // assumes there is a 512-byte header
	if ((fp = fopen (fn, "rb")) == NULL) {
		memset (h, 0, sizeof (struct headerdata));
		debug0 ("Failed to open %s to load header data!", fn);
		return;
	}
	fseek (fp, 65472+21+h->hadd, SEEK_SET);
	b = (fgetc (fp) & 0x0F);
	if (b == 1) {
		fseek (fp, 65472+28+h->hadd, SEEK_SET);
		fread (c, sizeof (word), 2, fp);
	}
	if ((b == 1 && (c[0] | c[1]) == 0xFFFF || forcerom == FORCEHIROM) && forcerom != FORCELOROM) {
		h->hirom = true;
		fseek (fp, 65530+h->hadd, SEEK_SET);
		fread (&h->nmivector, 2, 1, fp);
		fread (&h->resetvector, 2, 1, fp);
		fseek (fp, 65472+h->hadd, SEEK_SET);
	} else {
		h->hirom = false;
		fseek (fp, 32762+h->hadd, SEEK_SET);
		fread (&h->nmivector, 2, 1, fp);
		fread (&h->resetvector, 2, 1, fp);
		fseek (fp, 32704+h->hadd, SEEK_SET);
	}
	fread (h->name, 21, 1, fp);
	if (((b = fgetc (fp)) & 0x30) == 0x30) h->fastrom = true;
	else h->fastrom = false;
	fgetc (fp); // Ignore 'ROM type' byte
	h->romsize = 1 << (fgetc (fp) - 7);
	if (h->romsize == 0) h->romsize = 1;
	debug0 ("  Loaded h->romsize = %d", h->romsize);
	h->sramsize = 1 << (fgetc (fp) + 3);
	h->country = fgetc (fp);
	h->license = fgetc (fp);
	h->version = fgetc (fp);
	if (h->resetvector == 0) h->resetvector = 0x8000;
	fclose (fp);
}

void loadrom (char *fn, boolean gui, int forcerom)
{
	int x, c, pos;
	FILE *fp;
	struct headerdata h;
	boolean splitfile = false;

	debug0 ("Loadrom() entered. (Current ROM %s; New ROM %s)", curgs->filename, fn);
	c = fn [findfirstch (fn, '.') + 1];
	if (c == '1' || c == 'A' || c == 'a')
		splitfile = true;
	if ((fp = fopen (fn, "rb")) == NULL) {
		if (gui)
			messagebox (fn, "There was an error opening the file!");
		else {
			printf ("There was an error opening %s!\n", fn);
			getch ();
		}
		return;
	}
	debug0 ("  curgs->filename %s after fopen", curgs->filename);
	loadromheaderdata (fn, &h, forcerom);
	debug0 ("  curgs->filename %s after loadromheaderdata", curgs->filename);
	if (gui) {
		storewinsettings ();
		debug0 ("  curgs->filename %s After storewinsettings", curgs->filename);
		for (x = 1; x < NUMWINDOWS; x++) {
			killwindow (x);
		}
		mbc [2].hidden = false;
		mbc [1].hidden = true;
		sprintf (mbc[0].text[0], "%s", fn);
		strcpy (messagewindow.title, "Loading, please wait...");
		debug0 ("  curgs->filename %s After strcpy. h.romsize = %d", curgs->filename, h.romsize);
		mbc[2].max = h.romsize;
		showwindow (&messagewindow);
		debug0 ("  curgs->filename %s After showwindow", curgs->filename);
	} else {
		printf ("Loading ROM: %s\n", fn);
	}
	debug0 ("Before unloadrom");
	unloadcurrentrom ();
	debug0 ("got here! after unloadcurrentrom");
	if (forcerom == FORCELOROM)                 rom = (byte *) malloc (3145728); // 3 MB or up to 24 Mbits
	else if (forcerom == FORCEHIROM || h.hirom) rom = (byte *) malloc (4194304); // 4 MB or up to 32 Mbits
	else                                        rom = (byte *) malloc (3145728); // 3 MB or up to 24 Mbits
	romloaded = true;
	debug0 ("got here! after malloc");
	if (rom == NULL) {
		if (gui) {
			messagebox ("ROM Load Error", "Out of memory!!!");
		} else {
			printf ("Out of memory!!! Can't load ROM!!!\n");
			getch ();
		}
		fclose (fp);
		return;
	}
	fseek (fp, h.hadd, SEEK_SET);
	debug0 ("got here! after fseek h.hadd");
	for (pos = 0; !feof (fp); pos += 32768) {
		x = fread (rom + pos, 1, 32768, fp);
		if ((x < 32768 || feof (fp)) && splitfile) {
			c = findfirstch (fn, '.') + 1;
			fn[c]++;
			if ((fp = fopen (fn, "rb")) == NULL)
				break;
			fread (rom + pos + x, 1, 32768 - x, fp);
		}
		if (pos % (1024*128) == 0) {
			debug0 ("Loaded Megabit #%d", pos / (1024*128));
			if (gui) {
				mbc[2].choice = (pos / (1024*128));
				showgui ();
			} else {
				printf ("Loaded Megabit #%d\n", pos / (1024*128));
			}
		}
	}
	fclose (fp);
	createromentry (fn);
	curheader = h;
	debug0 ("Before reset");
	resetsystem (false); // not a warm boot - reset everything
	loadsram ();
	if (gui) {
		
		killwindow (&messagewindow);
		debug0 ("b4 applywinsettings");
		applywinsettings ();
	}
	for (x = 0; x < curgs->patches; x++) {
		if (curgs->cheat[x]->enabled) {
			curgs->cheat[x]->enabled = false;
			togglecheat (x); // Get original value... after this call, cheat is toggled on
		}
	}
	debug0 ("Finished loading ROM");
}

void unloadcurrentrom ()
{
	if (romloaded) {
		savesram ();
		resetsystem (true);
		free (rom);
		dealloc_autobacktrack();
		romloaded = false;
	}
}

