/*  SCREEN.CPP - SNES Screen Renderer
	Color depths (cdepth): 0=4 colors 1=16 colors 3=256 colors.
*/

#include "Common.hpp"
#include "GrEngine.hpp"
#include "CPU.hpp"
#include "FileMan.hpp"
#include "GUI.hpp"
#include "SPC700.hpp"
#include "Screen.hpp"
#include <ASSERT.H>

int debugscreen;

int memdumpoffset = 0, vramdumpoffset = 0, oamdumpoffset = 0, cgramdumpoffset = 0;
int vramblitoffset = 0, vramblitcdepth = 0;
int debug_sprmin = 0, debug_sprmax = 127;

char vramcmpfn[40] = "";
byte debug_vram[65536];

union cachetile ctile [4][4096]; // 1MB!
byte cstatus [4096];
byte csolid [2][4096], cempty [2][4096];  // 8 bits, 1 for each scanline of the tile.
	// Two sets 0=no vertical flip, 1=vertical flip

struct cachetilemap ctmap [4][128][128][2], sprmap [128][8][8];
	// 1MB! + 64K
	// here's the deal with the ctmap (cached tile map): 4 BGs*128 max height*
	// 128 max width*2 priorities*8 bytes=1MB.  This stores a more useful version of the
	// tile maps.  The unused priority entry has it's "empty" byte set to 0xFF, so nothing
	// is drawn.  Every scanline, one line of this cached tile map is built from the tile
	// map data in VRAM (except that the same line will not be built two scanlines in a row,
	// so it may be up to 8 lines between line copies.)
	// Now, if the BG isn't large enough horizontally to fill 128 elements, it is copied so
	// that it does.  For example if it's 32 tiles wide, those 32 built tiles are copied
	// 4 times to fill 128 elements.

byte spr8x8size [128]; // Set to 0 when cache entry is invalidated.

struct bginfo bg[4], *curbg;

union cachetile *translatetile (int taddress, byte cdepth, int flip)
{
	byte *tile = (byte*) &(ctile[flip][taddress]), *o_tile = tile;
	byte *addr = vram + (taddress << 4), *o_addr = addr, temp;
	signed char x, y, incaddr;
	boolean issolid = true, vflip = flip >> 1;
	byte *solidflags = &csolid[vflip][taddress], *emptyflags = &cempty[vflip][taddress];

	taddress &= 4095;
	if (flip & 2) {
		incaddr = -2;
		addr += 14;
	} else
		incaddr = 2;
	//debug0 ("  Caching tile; cstatus=%X cdepth=%d address=%d taddress=%d", cstatus[taddress], cdepth, taddress << 4, taddress);
	*solidflags = 0;
	*emptyflags = 0;
	switch (cdepth) {
	case 0:
		for (y = 0; y < 8; y++, tile += 8, addr += incaddr) {
			x = 7;
			do {
				tile[7-x] = ((addr[0] & (1<<x)) ?1:0) | (addr[1] & (1<<x) ?2:0);
				x--;
			} while (x >= 0);
			if ((addr[0] | addr[1]) == 0xFF) *solidflags |= (1 << y);
			else if ((addr[0] | addr[1]) == 0) *emptyflags |= (1 << y);
		}
		break;
	case 1:
		for (y = 0; y < 8; y++, tile += 8, addr += incaddr) {
			x = 7;
			do {
				tile[7-x] = (addr[0] & (1<<x) ?1:0) | (addr[1] & (1<<x) ?2:0) | (addr[16] & (1<<x) ?4:0) | (addr[17] & (1<<x) ?8:0);
				x--;
			} while (x >= 0);
			temp = addr[0] | addr[1] | addr[16] | addr[17];
			if (temp == 0xFF) *solidflags |= (1 << y);
			else if (temp == 0) *emptyflags |= (1 << y);
		}
		break;
	case 3:
		for (y = 0; y < 8; y++, tile += 8, addr += incaddr) {
			x = 7;
			do {
				tile[7-x] = (addr[0] & (1<<x) ?1:0) | (addr[1] & (1<<x) ?2:0) | (addr[16] & (1<<x) ?4:0) | (addr[17] & (1<<x) ?8:0) \
						  | (addr[32]& (1<<x)?16:0) | (addr[33]& (1<<x)?32:0) | (addr[48] & (1<<x)?64:0) | (addr[49] &(1<<x)?128:0);
				x--;
			} while (x >= 0);
			temp = addr[0] | addr[1] | addr[16] | addr[17] | addr[32] | addr[33] | addr[48] | addr[49];
			if (temp == 0xFF) *solidflags |= (1 << y);
			else if (temp == 0) *emptyflags |= (1 << y);
		}
		break;
	}
	cstatus [taddress] = cdepth | (cstatus [taddress] & 0xF0) | (16 << flip);
	tile = o_tile;
	if (flip & 1) {
		for (y = 0; y < 8; y++) {
			temp = tile[7]; tile[7] = tile[0]; tile[0] = temp;
			temp = tile[6]; tile[6] = tile[1]; tile[1] = temp;
			temp = tile[5]; tile[5] = tile[2]; tile[2] = temp;
			temp = tile[4]; tile[4] = tile[3]; tile[3] = temp;
			tile += 8;
		}
	}
	/*
	if ((taddress<<4) >= 0xC000 && (taddress<<4) < 0xC200) { // second row first ball
		for (y=0;y<8;y++) {
			debug0 ("    %d %02X %02X %02X %02X %02X %02X %02X %02X   P0:%04X", y,
			ctile[taddress].pixel[y*8], ctile[taddress].pixel[y*8+1],
			ctile[taddress].pixel[y*8+2], ctile[taddress].pixel[y*8+3],
			ctile[taddress].pixel[y*8+4], ctile[taddress].pixel[y*8+5],
			ctile[taddress].pixel[y*8+6], ctile[taddress].pixel[y*8+7],
			*((word*)&o_addr[y*2]));
		}
	}
	*/
	return &ctile [flip][taddress];
}

inline union cachetile *getcachetile (int taddress, byte cdepth, int flip)
{ // COLOR DEPTH: 0=2 bit 1=4 bit 3=8 bit
	if ((cstatus[taddress] & (16 << flip)) && (cstatus[taddress] & 0x3) == cdepth) {
		return &ctile [flip][taddress];
	} else {
		return translatetile (taddress, cdepth, flip);
	}
}

byte *screen, scanlines;
	// screen points to current scanline; scanlines = 224 or 239.

struct snesscreenmode {
	int bgs;
	byte cdepth[4];
} snesmode [8] = {
	{ 4, 0, 0, 0, 0 },
	{ 3, 1, 1, 0 },
	{ 2, 1, 1 },
	{ 2, 3, 1 },
	{ 2, 3, 0 },
	{ 2, 1, 0 },
	{ 1, 1 },
	{ 0 }
};

void resetscreenframe ()
{       // To be called at scanline 0...
	int x;

	screen = vs + 256*8;
	for (x = 0; x < 4; x++) {
		getbgframeinfo (x);
		bg[x].lastbgrow = -1;
	}
	for (x = 0; x < 256; x++) {
		setsnespalette (x, ((word*)cgram) [x], (*REG2100 & 0xF));
	}
	if ((*REG212C | *REG212D | curgs->bgs) & 0x10) {
		buildspritetable ();
		oamposchange = true;
	}
}

void getbgframeinfo (int plane)
{
	bg[plane].bgnum = plane + 1;
}

void getbglineinfo (int plane)
{
	bg[plane].cdepth = snesmode[*REG2105 & 7].cdepth[plane];
	bg[plane].tcharshift = (bg[plane].cdepth == 3 ? 2 : bg[plane].cdepth);
	if (*REG2105 & (16 << plane)) { // 16x16 tiles
		bg[plane].tile16x16 = true;
	} else {
		bg[plane].tile16x16 = false;
	}
	if (*REG2107(plane) & 2) { // SC Height
		bg[plane].height64 = true;
		//bg[plane].pixheight = (512 << bg[plane].tile16x16);
		//bg[plane].pixheightand = (bg[plane].tile16x16 ? 0x3FF : 0x1FF);
		bg[plane].height8x8 = (64 << bg[plane].tile16x16);
	} else {
		bg[plane].height64 = false;
		//bg[plane].pixheight = (256 << bg[plane].tile16x16);
		//bg[plane].pixheightand = (bg[plane].tile16x16 ? 0x1FF : 0xFF);
		bg[plane].height8x8 = (32 << bg[plane].tile16x16);
	}
	if (*(REG2107(plane)) & 1) { // SC Width
		bg[plane].width64 = true;
		//bg[plane].pixwidth = (512 << bg[plane].tile16x16);
		//bg[plane].pixwidthand = (bg[plane].tile16x16 ? 0x3FF : 0x1FF);
		//bg[plane].pixwidthshift = (9 + bg[plane].tile16x16);
		bg[plane].width8x8 = (64 << bg[plane].tile16x16);
	} else {
		bg[plane].width64 = false;
		//bg[plane].pixwidth = (256 << bg[plane].tile16x16);
		//bg[plane].pixwidthand = (bg[plane].tile16x16 ? 0x1FF : 0xFF);
		//bg[plane].pixwidthshift = (8 + bg[plane].tile16x16);
		bg[plane].width8x8 = (32 << bg[plane].tile16x16);
	}
	bg[plane].width8x8and = bg[plane].width8x8 - 1;
	bg[plane].height8x8and = bg[plane].height8x8 - 1;
	bg[plane].tilemap = vram + ((*REG2107(plane) & 0x7C) << 9); // Correct in demos
	bg[plane].tcharbase = ((*REG210B >> (plane << 2)) & 7) << 9; // Correct in demos
	bg[plane].scrollx = (state.scrollreg[plane << 1].scroll   & 0x7FF);
	bg[plane].scrolly = (state.scrollreg[(plane<<1)+1].scroll & 0x7FF);// & bg[plane].pixelheightmask);
	if ((*REG2105 & 7) == 0) // Mode 0
		bg[plane].cgoffs = (plane << 5);
	else
		bg[plane].cgoffs = 0;
}

void dobglinedraw (int plane, int on, boolean priority)
{
	int row, offsy;

	if (snesmode[*REG2105&7].bgs > plane && (on & (1 << plane)) && (curgs->bgs & (1 << plane))) {
		curbg = &bg[plane];
		row = ((curbg->scrolly >> 3) & curbg->height8x8and);
		offsy = curbg->scrolly & 7;
		if (row != curbg->lastbgrow) {
			updatetilemaprow (row); // uses curbg
			curbg->lastbgrow = row;
		}
		blitbgline (row, offsy, (curbg->scrollx + 7) >> 3, screen + ((8 - curbg->scrollx) & 7), priority);
		// row, offsy, col, scrstart, priority
	}
	if ((*REG2105&7) == 7) {
		blitmode7line ();
	}
}

void blitmode7line (void)
{
	/*int pixx, x, y;

	for (pixx = 0; pixx < 256; pixx++) {
		x = pixx << 8;
		y = screen_scanline << 8;
	}*/
}

void drawnextscanline ()
{
	int x, on;

	screen = vs + (screen_scanline+8) * 256;

	if (*REG2100 & 0x80) // Screen off
		return;
	debug0 ("scanline %d", screen_scanline);
	for (x = 0; x < snesmode[*REG2105&7].bgs; x++) {
		getbglineinfo (x);
		// Now, offset line-specific information...
		bg[x].scrolly += screen_scanline;
		debug0 ("getbginfo %d",x);
	}
	if (oamposchange) {
		debug0 ("Build sprlinelist");
		buildsprlinelist ();
	}

	on = *REG212D & ~(*REG212C); // Draw sub screens
	debug0 ("subscreen on %X",on);
	if ((*REG2105 & 15) == 8 + 1) {
		dobglinedraw (3, on, 0); dobglinedraw (2, on, 0);
		dobglinedraw (3, on, 1); dobglinedraw (1, on, 0);
		dobglinedraw (0, on, 0); dobglinedraw (1, on, 1);
		dobglinedraw (0, on, 1); dobglinedraw (2, on, 1);
	} else {
		dobglinedraw (3, on, 0); dobglinedraw (2, on, 0);
		dobglinedraw (3, on, 1); dobglinedraw (2, on, 1);
		dobglinedraw (1, on, 0); dobglinedraw (0, on, 0);
		dobglinedraw (1, on, 1); dobglinedraw (0, on, 1);
	}
	on = *REG212C;
	debug0 ("mainscreen on %X",on);
	if ((*REG2105 & 15) == 8 + 1) {
		dobglinedraw (3, on, 0);
		dobglinedraw (2, on, 0);
		if (on&0x10) dospritelinedraw (0);
		dobglinedraw (3, on, 1);
		if (on&0x10) dospritelinedraw (1);
		dobglinedraw (1, on, 0);
		dobglinedraw (0, on, 0);
		dobglinedraw (1, on, 1);
		if (on&0x10) dospritelinedraw (2);
		dobglinedraw (0, on, 1);
		if (on&0x10) dospritelinedraw (3);
		dobglinedraw (2, on, 1);
	} else {
		dobglinedraw (3, on, 0);
		dobglinedraw (2, on, 0);
		if (on&0x10) dospritelinedraw (0);
		dobglinedraw (3, on, 1);
		dobglinedraw (2, on, 1);
		if (on&0x10) dospritelinedraw (1);
		dobglinedraw (1, on, 0);
		dobglinedraw (0, on, 0);
		if (on&0x10) dospritelinedraw (2);
		dobglinedraw (1, on, 1);
		dobglinedraw (0, on, 1);
		if (on&0x10) dospritelinedraw (3);
	}
}

void rendersnesscreen ()
{
	if (debugmode) {
		drawdebug (debugscreen);
	} else {
		doframedraw ();
	}
}

void updatetilemap64_16x16 (byte *tilemap, struct cachetilemap *ctmap, int tcharbase, boolean secondrow)
{	// version for 16x16 tile maps.
	int x, tnum, prio;

	if (secondrow) // only difference between 1st & 2nd row is +2 chars...
		tcharbase += (16 << curbg->tcharshift);
	for (x = 32; x > 0; x--) {
		tnum = tcharbase + ((*(word*)tilemap & 0x3FF) << curbg->tcharshift);
		prio = (tilemap[1] >> 5) & 1;
		ctmap[prio].tile = getcachetile (tnum, curbg->cdepth, tilemap[1] >> 6);
		ctmap[prio].empty = cempty [tilemap[1]>>7][tnum];
		ctmap[prio].solid = csolid [tilemap[1]>>7][tnum];
		ctmap[prio].pal_or = curbg->cgoffs | ((tilemap[1] & 0x1C) << (curbg->cdepth<<1));
		tnum += 1 << curbg->tcharshift;
		ctmap[prio+2].tile = getcachetile (tnum, curbg->cdepth, tilemap[1] >> 6);
		ctmap[prio+2].empty = cempty [tilemap[1]>>7][tnum];
		ctmap[prio+2].solid = csolid [tilemap[1]>>7][tnum];
		ctmap[prio+2].pal_or = curbg->cgoffs | ((tilemap[1] & 0x1C) << (curbg->cdepth<<1));
		prio = !prio;
		ctmap[prio].tile = NULL;
		ctmap[prio].empty = 0xFF;
		ctmap[prio].solid = ctmap[prio].pal_or = ctmap[prio].dummy = 0;
		ctmap[prio+2].tile = NULL;
		ctmap[prio+2].empty = 0xFF;
		ctmap[prio+2].solid = ctmap[prio].pal_or = ctmap[prio].dummy = 0;
		tilemap += 2;
		ctmap += 4;
	}
}

void updatetilemap32 (byte *tilemap, struct cachetilemap *ctmap, int tcharbase)
{
	int x, tnum, prio;

	for (x = 32; x > 0; x--) {
		tnum = tcharbase + ((*(word*)tilemap & 0x3FF) << curbg->tcharshift);
		prio = (tilemap[1] >> 5) & 1;
		ctmap[prio].tile = getcachetile (tnum, curbg->cdepth, tilemap[1] >> 6);
		ctmap[prio].empty = cempty [tilemap[1]>>7][tnum];
		ctmap[prio].solid = csolid [tilemap[1]>>7][tnum];
		ctmap[prio].pal_or = curbg->cgoffs | ((tilemap[1] & 0x1C) << (curbg->cdepth << 1));
		prio = !prio;
		ctmap[prio].tile = NULL;
		ctmap[prio].empty = 0xFF;
		ctmap[prio].solid = ctmap[prio].pal_or = ctmap[prio].dummy = 0;
		tilemap += 2;
		ctmap += 2;
	}
}

void updatetilemaprow (int row)
{	// uses curbg
	byte *tilemap = curbg->tilemap;
	struct cachetilemap *ctilemap = (struct cachetilemap *) ctmap[curbg->bgnum-1][row];

	if (row >= (32<<curbg->tile16x16)) {
		row -= (32<<curbg->tile16x16);
		tilemap += ((32*32*2) << curbg->width64);
	}
	if (!curbg->tile16x16) { // 8x8 tiles
		updatetilemap32 (tilemap + (row << 6), ctilemap, curbg->tcharbase);
		if (curbg->width64) { // 64 wide
			updatetilemap32 (tilemap + 32*32*2 + (row << 6), ctilemap + 32*2, curbg->tcharbase);
		} else {
			memcpy (ctilemap + 32*2, ctilemap, sizeof(cachetilemap)*32*2);
		}
		memcpy (ctilemap + 64*2, ctilemap, sizeof(cachetilemap)*64*2);
	} else { // 16x16 tiles
		updatetilemap64_16x16 (tilemap + ((row >> 1) << 6), ctilemap, curbg->tcharbase, row & 1);
		if (curbg->width64) { // 64 16x16 tiles wide
			updatetilemap64_16x16 (tilemap + 32*32*2 + ((row >> 1) << 6), ctilemap + 64*2, curbg->tcharbase, row & 1);
		} else {
			memcpy (ctilemap + 64*2, ctilemap, sizeof(cachetilemap)*64*2);
		}
	}
}

void blitbgline (int row, int offsy, int col, byte *scrstart, boolean priority)
{	// uses curbg
	struct cachetilemap *ctilemap = &ctmap[curbg->bgnum-1][row][col][priority];
	int x, flagand = (1 << offsy), pixoffsy = (offsy << 3);
	dword dwpal_or;
	byte *pixels;

	for (x = 32; x != 0; scrstart += 8, ctilemap += 2, col++, x--) {
		if (col & 0xFFFFFF80) { // 0 <= Column < 128
			col -= 128;
			ctilemap -= 256; // 128*2
		}
		if (ctilemap->empty & flagand)
			continue;
		else if (ctilemap->solid & flagand) {
			dwpal_or = (ctilemap->pal_or << 24) | (ctilemap->pal_or << 16) | (ctilemap->pal_or << 8) | ctilemap->pal_or;
			((dword*)scrstart)[0] = ctilemap->tile->dwpixel[(offsy << 1)] | dwpal_or;
			((dword*)scrstart)[1] = ctilemap->tile->dwpixel[(offsy << 1) + 1] | dwpal_or;
			continue;
		}
		pixels = &ctilemap->tile->pixel[pixoffsy];
		if (pixels[0]) scrstart[0] = pixels[0] | ctilemap->pal_or;
		if (pixels[1]) scrstart[1] = pixels[1] | ctilemap->pal_or;
		if (pixels[2]) scrstart[2] = pixels[2] | ctilemap->pal_or;
		if (pixels[3]) scrstart[3] = pixels[3] | ctilemap->pal_or;
		if (pixels[4]) scrstart[4] = pixels[4] | ctilemap->pal_or;
		if (pixels[5]) scrstart[5] = pixels[5] | ctilemap->pal_or;
		if (pixels[6]) scrstart[6] = pixels[6] | ctilemap->pal_or;
		if (pixels[7]) scrstart[7] = pixels[7] | ctilemap->pal_or;
	}
}

inline int getobj8x8size (int bits)
{
	switch ((bits & 2) | (*REG2101 & 0xE0)) {
	case 0x00: case 0x20: case 0x40: return 1;
	case 0x02: case 0x60: case 0x80: return 2;
	case 0xA0: case 0x22: case 0x62: return 4;
	case 0x42: case 0x82: case 0xA2: return 8;
	}
	return 1;	// Size unknown
}

void buildspritetable ()
{	// Called at the beginning of every frame
	int obj, bits, tchars, flip, x, y, i, incy;
	byte pal_or;
    int tcharbase = ((*REG2101 & 3) << 10); // Char Address/16
	struct cachetilemap *cursprmap;
	static struct cachetilemap dummymap = { NULL, 0, 0xFF, 0 };

	for (obj = 0; obj < 128; obj++) {
		/*if (spr8x8size[obj] != 0) {
			tchars = tcharbase + ((*((word*)&oam[(obj << 2) + 2]) & 0x1FF) << 1);
			flip = oam [(obj << 2)+3] >> 6;
			for (y = 0; y < spr8x8size[obj]; y++) {
				for (x = 0; x < spr8x8size[obj]; x++) {
					//cursprmap[(y << 3) + x].tile
					getcachetile (tchars + (x << 1), 1, flip);
				}
				tchars += 32;
			}
			continue;
		}*/
		bits = (oam [512 + (obj >> 2)] & (3 << ( (obj & 3) << 1 ) )) >> ( (obj & 3) << 1);
		spr8x8size[obj] = getobj8x8size (bits);
		flip = oam [(obj << 2)+3] >> 6;
		tchars = tcharbase + ((*((word*)&oam[(obj << 2) + 2]) & 0x1FF) << 1);
		pal_or = ((oam[(obj << 2) + 3] & 0xE) << 3) | 0x80;
		cursprmap = (struct cachetilemap *) sprmap[obj];
		if (spr8x8size[obj] == 1) { //8x8=easy
			cursprmap[0].tile = getcachetile (tchars, 1, flip);
			cursprmap[0].pal_or = pal_or;
			cursprmap[0].empty = cempty [flip>>1][tchars];
			cursprmap[0].solid = csolid [flip>>1][tchars];
			cursprmap[1] = cursprmap[2] = cursprmap[3] = cursprmap[4] =\
			cursprmap[5] = cursprmap[6] = cursprmap[7] = dummymap;
			continue;
		}
		if (flip & 2) {
			incy = -1; y = spr8x8size[obj]-1;
		} else {
			incy = 1; y = 0;
		}
		for (; y < spr8x8size[obj] && y >= 0; y += incy) {
			for (x = 0; x < spr8x8size[obj]; x++) {
				if (flip & 1) i = (y << 3) + (spr8x8size[obj]-1-x);
				else i = (y << 3) + x;
				cursprmap[i].tile = getcachetile (tchars + (x << 1), 1, flip);
				cursprmap[i].pal_or = pal_or;
				cursprmap[i].empty = cempty [flip>>1][tchars + (x << 1)];
				cursprmap[i].solid = csolid [flip>>1][tchars + (x << 1)];
			}
			for (; x < 8; x++)
				cursprmap[(y << 3) + x] = dummymap; // should be unneccesary?
			tchars += 32;
		}
	}
}

boolean oamposchange;
byte sprhere [256][32];
byte nsprhere [256];

void buildsprlinelist ()
{
	int bits, xpos, ypos, size, obj, y;

	memset (nsprhere, 0, sizeof (nsprhere));
	for (obj = (debugmode ? debug_sprmax : 127); obj >= (debugmode ? debug_sprmin : 0); obj--) {
		bits = (oam [512 + (obj >> 2)] & (3 << ( (obj & 3) << 1 ) )) >> ( (obj & 3) << 1);
		xpos = oam [(obj << 2)] + ((bits & 1) << 8);
		ypos = oam [(obj << 2) + 1];
		size = spr8x8size[obj] << 3;
		if (xpos >= 256 && xpos < 512+7-size)
			continue;
		for (y = ypos; y < ypos+size; y++) {
			if (nsprhere[y & 0xFF] >= 32)
				continue;
			sprhere [y & 0xFF][nsprhere[y & 0xFF]] = obj;
			nsprhere[y & 0xFF]++;
		}
	}
	oamposchange = false;
}

/*void dospritelinedraw (byte priority)
{
	int bits, xpos, ypos, size, row, col, pixoffsy, numtiles, obj;

	if (!(curgs->bgs & 0x10))
		return;
	for (obj = 127; obj >= 0; obj--) {
		if (((oam [(obj << 2) + 3] & 0x30) >> 4) != priority)
			continue;
		bits = (oam [512 + (obj >> 2)] & (3 << ( (obj & 3) << 1 ) )) >> ( (obj & 3) << 1);
		xpos = oam [(obj << 2)] + ((bits & 1) << 8);
		ypos = oam [(obj << 2) + 1];
		size = spr8x8size[obj] << 3;
		if (xpos > 256 && xpos < 512+7-size)
			continue;
		if ((screen_scanline >= ypos+size || screen_scanline < ypos) &&\
			!(ypos+size >= 256 && ((ypos+size)&0xFF) > ypos))
			continue;
		pixoffsy = (screen_scanline - ypos) & 0xFF;
		row = pixoffsy >> 3;
		pixoffsy &= 7;
		numtiles = (256+7-xpos) >> 3;
		if (numtiles > spr8x8size[obj])
			numtiles = spr8x8size[obj];
		col = xpos >= 256 ? (512+7-xpos) >> 3 : 0;
		blitspriteline (&sprmap[obj][row][col], numtiles, pixoffsy, screen + ((xpos+(col<<3))&0xFF));
	}
}*/

void dospritelinedraw (byte priority)
{
	int bits, xpos, ypos, size, row, col, pixoffsy, numtiles, obj, obji;

	if (!(curgs->bgs & 0x10))
		return;
	for (obji = 0; obji < nsprhere[screen_scanline]; obji++) {
		obj = sprhere[screen_scanline][obji];
		if (((oam [(obj << 2) + 3] & 0x30) >> 4) != priority)
			continue;
		bits = (oam [512 + (obj >> 2)] & (3 << ( (obj & 3) << 1 ) )) >> ( (obj & 3) << 1);
		xpos = oam [(obj << 2)] + ((bits & 1) << 8);
		ypos = oam [(obj << 2) + 1];
		size = spr8x8size[obj] << 3;

		pixoffsy = (screen_scanline - ypos) & 0xFF;
		row = pixoffsy >> 3;
		pixoffsy &= 7;
		numtiles = (256+7-xpos) >> 3;
		if (numtiles > spr8x8size[obj])
			numtiles = spr8x8size[obj];
		col = xpos >= 256 ? (512+7-xpos) >> 3 : 0;
		blitspriteline (&sprmap[obj][row][col], numtiles, pixoffsy, screen + ((xpos+(col<<3))&0xFF));
	}
}

void blitspriteline (struct cachetilemap *ctmap, int numtiles, int yoffs, byte *scrstart)
{
	int flagand = (1 << yoffs), pixoffsy = (yoffs << 3);
	byte *pixels;
	dword dwpal_or;

	for (;numtiles > 0; numtiles--, scrstart += 8, ctmap++) {
		if (ctmap->empty & flagand) {
			continue;
		}
		else if (ctmap->solid & flagand) {
			dwpal_or = (ctmap->pal_or << 24) | (ctmap->pal_or << 16) | (ctmap->pal_or << 8) | ctmap->pal_or;
			((dword*)scrstart)[0] = ctmap->tile->dwpixel[(yoffs << 1)] | dwpal_or;
			((dword*)scrstart)[1] = ctmap->tile->dwpixel[(yoffs << 1) + 1] | dwpal_or;
			continue;
		}
		pixels = &ctmap->tile->pixel[pixoffsy];
		if (pixels[0]) scrstart[0] = pixels[0] | ctmap->pal_or;
		if (pixels[1]) scrstart[1] = pixels[1] | ctmap->pal_or;
		if (pixels[2]) scrstart[2] = pixels[2] | ctmap->pal_or;
		if (pixels[3]) scrstart[3] = pixels[3] | ctmap->pal_or;
		if (pixels[4]) scrstart[4] = pixels[4] | ctmap->pal_or;
		if (pixels[5]) scrstart[5] = pixels[5] | ctmap->pal_or;
		if (pixels[6]) scrstart[6] = pixels[6] | ctmap->pal_or;
		if (pixels[7]) scrstart[7] = pixels[7] | ctmap->pal_or;
	}
}

void doframedraw()
{
	int temp = screen_scanline, y;

	clearscreen (0);
	resetscreenframe ();
	screen_scanline = 0;
	for (y = 0; y < 4; y++) {
		getbgframeinfo (y);
		getbglineinfo (y);
		setcolor (0);
		curbg = &bg[y];
		printf8 (201, 51+(curbg->bgnum*10), "%d(%d,%d)", curbg->bgnum, curbg->scrollx, curbg->scrolly);
		setcolor (messagecolor);
		printf8 (200, 50+(curbg->bgnum*10), "%d(%d,%d)", curbg->bgnum, curbg->scrollx, curbg->scrolly);
		debug0 ("bgnum=%d scrollstartx=%d scrollstarty=%d tile16x16=%d cgoffs=%d cdepth=%d width64=%d height=%d",
			curbg->bgnum, curbg->scrollx, curbg->scrolly, curbg->tile16x16, curbg->cgoffs, curbg->cdepth, curbg->width64, curbg->height64);
		debug0 ("BG: width8x8=%d width8x8and=%d height8x8=%d height8x8and=%d tchars=%d tcharshift=%d",
			curbg->width8x8, curbg->width8x8and, curbg->height8x8, curbg->height8x8and, curbg->tcharbase, curbg->tcharshift);
	}
	while (screen_scanline < scans_before_vbl()) {
		drawnextscanline ();
		//copyscreen ();
		screen_scanline++;
	}
	screen_scanline = temp;
}






















#define DEBUGLINES 27

struct debugline {
	char text [55];
	boolean showregs;
	struct regpack reg;
	char asterisk;
	int color;
} line [DEBUGLINES];

int debuglines = 1;

void deletefirstdebugline ()
{
	int x;

	for (x = 0; x < debuglines; x++) {
		line[x] = line[x + 1];
	}
	debuglines--;
}

void adddebugline (char *text, boolean showregs, char asterisk)
{
	dword opbyte;

	debug0 ("debug A%0.4X X%0.4X Y%0.4X P%0.2X PC$%X OP$%X %c%s", reg.A, reg.X, reg.Y, reg.P, reg.PC, *SNESMEM(reg.PC), asterisk, text);
	if (debuglines == DEBUGLINES - 1)
		deletefirstdebugline ();
	line[debuglines].reg = reg;
	//line[debuglines].reg.PC = PC; line[debuglines].reg.DBR = DBR;
	//line[debuglines].reg.Y = Y; line[debuglines].reg.X = X;
	//line[debuglines].reg.D = D; line[debuglines].reg.A = A;
	//line[debuglines].reg.P = P; line[debuglines].reg.E = E;
	//line[debuglines].reg.S = S;
	line[debuglines].showregs = showregs;
	line[debuglines].asterisk = asterisk;
	line[debuglines].color = 127;
	strcpy (line[debuglines].text, text);
	if (showregs && debuglines != 0 && line[debuglines-1].asterisk == ' ') {
		if (line[debuglines-1].reg.PC >= reg.PC)
			line[debuglines-1].asterisk = 30;
		opbyte = *SNESMEM (line[debuglines-1].reg.PC);
		if (line[debuglines-1].reg.PC + opcodelist[opbyte].bytes + 1 < reg.PC)
			line[debuglines-1].asterisk = 31;
	}
	debuglines++;
}

#define DEBUG_SETREGCOLOR(index,field) setcolor ((index) == debuglines - 1 ? 45 : (line[index].reg.field) == (line[index+1].reg.field) ? 15 : 123)

void drawdebugline (int x, boolean regs2)
{
	int b;

	setcolor (line[x].color);
	printf8 (42, 27 + x*8, "%s", line[x].text);
	printf8 (0, 27 + x*8, "%c", line[x].asterisk);
	if (line[x].showregs) {
		setcolor (30);
		printf8 (8, 27+x*8, "%0.6X", line[x].reg.PC);
		DEBUG_SETREGCOLOR (x, A);
		if (line[x].reg.P & 0x20)
			drawbox (115, 26+x*8, 123, 33+x*8, 7);
		printf8 (115, 27+x*8, "%0.4X", line[x].reg.A);
		if (line[x].reg.P & 0x10) {
			drawbox (140, 26+x*8, 148, 33+x*8, 7);
			drawbox (165, 26+x*8, 173, 33+x*8, 7);
		}
		DEBUG_SETREGCOLOR (x, X);
		printf8 (140, 27+x*8, "%0.4X", line[x].reg.X);
		DEBUG_SETREGCOLOR (x, Y);
		printf8 (165, 27+x*8, "%0.4X", line[x].reg.Y);
		if (!regs2) {
			DEBUG_SETREGCOLOR (x, S);
			printf8 (190, 27+x*8, "%0.4X", line[x].reg.S);
			DEBUG_SETREGCOLOR (x, D);
			printf8 (215, 27+x*8, "%0.4X", line[x].reg.D);
			DEBUG_SETREGCOLOR (x, P);
			printf8 (240, 27+x*8, "%0.2X", line[x].reg.P);
		} else {
			DEBUG_SETREGCOLOR (x, DBR);
			printf8 (197, 27+x*8, "%0.2X", line[x].reg.DBR);
			for (b = 0; b < 8; b++) {
				setcolor (x == debuglines - 1 ? 45 : \
					((line[x].reg.P >> b) & 1) == ((line[x+1].reg.P >> b) & 1) ? 15 : 123);
				printf8 (211 + (7-b) * 5, 27+x*8, "%d", (line[x].reg.P >> b) & 1);
			}
			if (line[x].reg.D != 0) {
				DEBUG_SETREGCOLOR (x, D);
				printf8 (189, 27+x*8, "d", line[x].reg.D);
			}
		}
	} else if (line[x].asterisk == '*') {
		setcolor (21);
		printf8 (8, 27+x*8, "%0.6X", line[x].reg.PC);
	}
}

void drawmemdump (byte *area, char *s, int offs)
{
	int y, x;

	clearscreen ();
	setcolor (15);
	printf8 (0,0,"Debugger   Instr:%0.8d Scanline:%0.3d Cyc/scan:%0.3d/%0.3d", debug_instr, screen_scanline, scan_cycles, cycles_per_scan);
	printf8 (0,8,"Total frames:%0.4d               Memory dump view", total_frames);
	printf8 (0,16,"%s (Starting at offset %d)", s, offs);
	for (y = -2; y < 25; y++) {
		setcolor (174);
		if (y < 0) setcolor (165);
		printf8(0,42+y*8,"%0.6X", offs + y*12);
		setcolor (15);
		if (y < 0) setcolor (6);
		for (x = 0; x < 12; x++) {
			setcolor (15);
			if (debugscreen == 4 && vramcmpfn[0] != '\0') {
				if (y*12 + x + vramdumpoffset > 0x10000 || y*12 + x + vramdumpoffset < 0)
					continue;
				if (area [y * 12 + x] != debug_vram [y * 12 + x + vramdumpoffset]) {
					setcolor (24);
					printf8 (37+x*15, 43+y*8, "%0.2X", debug_vram [y * 12 + x + vramdumpoffset]);
					setcolor (124);
				}
			}
			if (((offs + y*12 + x) & 0xFF0000) == (offs & 0xFF0000))
				printf8 (35+x*15, 42+y*8, "%0.2X", area [y * 12 + x]);
		}
	}
}

void drawdebug (int which)
{
	int x, y, x2, y2;
	dword opdata;
	char *fmt;
	union cachetile *tile;

	if (which == 0) {
		doframedraw ();
	} else if (which == 1 || which == 2) {
		clearscreen ();
		setcolor (15);
		printf8 (0,0,"Debugger   Instr:%0.8d Scanline:%0.3d Cyc/scan:%0.3d/%0.3d", debug_instr, screen_scanline, scan_cycles, cycles_per_scan);
		printf8 (0,8,"Total frames:%0.4d               Type ? for command list", total_frames);
		opdata = *(dword*)SNESMEM (reg.PC);
		fmt = opcodelist[(byte)opdata].format;
		printf8 (0,16,"%c%c%c: %s", fmt[0], fmt[1], fmt[2], getdescr ((byte)opdata));
		setcolor (30);
		printf8 (8,26,"PC");
		printf8 (115,26,"A");
		printf8 (140,26,"X");
		printf8 (165,26,"Y");
		if (which == 2) {
			printf8 (197,26,"DB");
			printf8 (210,26,"NVMXDIZC");
		} else {
			printf8 (190,26,"S");
			printf8 (215,26,"D");
			printf8 (240,26,"P");
		}
		for (x = 1; x < debuglines; x++) {
			drawdebugline (x, which == 2 ? true : false);
		}
	} else if (which == 3) {
		drawmemdump (SNESMEM (memdumpoffset), "65c816 Memory Dump", memdumpoffset);
	} else if (which == 4) {
		drawmemdump (vram + vramdumpoffset, "VRAM Memory Dump", vramdumpoffset);
	} else if (which == 5) { // CGRAM dump
		drawmemdump (cgram + cgramdumpoffset, "CGRAM Memory Dump", cgramdumpoffset);
	} else if (which == 6) {
		drawmemdump (oam + oamdumpoffset, "OAM Memory Dump", oamdumpoffset);
	} else if (which == 7) {
		clearscreen ();
		setcolor (15);
		printf8 (0, 0, "Common Registers [ptrs: VRAM:%X OAM:%X CG:%X]", state.vrampointer, state.oampointer, state.colorpointer);
		printf8 (0, 8, "$2100($%X) b7:1=screen off b0-3:bright", *REG2100);
		printf8 (0,16, "$2105($%X) b0-2:mode b4-7:16x16 b3:BG3 priority", *REG2105);
		printf8 (0,24, "$2107($%X) b0-1:Sc.size b2-7:base(BG1) $%X", *REG2107(0), (*REG2107(0) & ~3) << 9);
		printf8 (0,32, "$2108($%X) b0-1:Sc.size b2-7:base(BG2) $%X", *REG2107(1), (*REG2107(1) & ~3) << 9);
		printf8 (0,40, "$2109($%X) b0-1:Sc.size b2-7:base(BG3) $%X", *REG2107(2), (*REG2107(2) & ~3) << 9);
		printf8 (0,48, "$210A($%X) b0-1:Sc.size b2-7:base(BG4) $%X", *REG2107(3), (*REG2107(3) & ~3) << 9);
		printf8 (0,56, "$210B($%X) b0-3:BG1 ch base $%X b4-7:BG2 $%X", *(byte*)REG210B, (*(byte*)REG210B & 15) << 13, (*(byte*)REG210B >> 4) << 13);
		printf8 (0,64, "$210C($%X) b0-3:BG3 ch base $%X b4-7:BG4 $%X", *(((byte*)REG210B) + 1), (*(((byte*)REG210B) + 1) & 15) << 13, (*(((byte*)REG210B) + 1) >> 4) << 13);
		printf8 (0,72, "$210D($%X/%d) $210E($%X/%d) BG1 Scroll H/V", state.scrollreg[0].scroll, state.scrollreg[0].hipointer, state.scrollreg[1].scroll, state.scrollreg[1].hipointer);
		printf8 (0,80, "$210F($%X/%d) $2110($%X/%d) BG1 Scroll H/V", state.scrollreg[2].scroll, state.scrollreg[2].hipointer, state.scrollreg[3].scroll, state.scrollreg[3].hipointer);
		printf8 (0,88, "$2111($%X/%d) $2112($%X/%d) BG1 Scroll H/V", state.scrollreg[4].scroll, state.scrollreg[4].hipointer, state.scrollreg[5].scroll, state.scrollreg[5].hipointer);
		printf8 (0,96, "$2113($%X/%d) $2114($%X/%d) BG1 Scroll H/V", state.scrollreg[6].scroll, state.scrollreg[6].hipointer, state.scrollreg[7].scroll, state.scrollreg[7].hipointer);
		printf8 (0,104, "$2115(%X) b7:inc@$2119? b0-1:Inc.rate b2-3:Full gr.", *REG2115);
		printf8 (0,112, "$2116(%X)*2 initial VRAM address", *REG2116);
		printf8 (0,120, "$2118(%X) VRAM data write", *REG2118);
		printf8 (0,128, "$212C(%X)/$212D(%X) Main/Sub screen designation", *REG212C, *REG212D);
		printf8 (0,136, "$211A(%X)/$211B(%X)/$211C(%X)/$211D(%X)/$211E(%X) Mode7", *REG211A, state.m7vars.cosx, state.m7vars.sinx, state.m7vars.cosy, state.m7vars.siny);
		printf8 (0,144, "$211F(%X/%d)/$2120(%X/%d) Setup,Cos&SinX,Cos&SinY,Center X/Y", state.m7vars.x, state.m7vars.hipointer[4], state.m7vars.y, state.m7vars.hipointer[5]);
		printf8 (0,152, "$2133(%X) b6:extbg b3:psd-512 b2:239-line b1:int.OBJ b0:int", *REG2133);
		printf8 (0,160, "$2180(%X=%X) WRAM byte 1: data byte 2-4: address", *(dword*)REG2181 & 0xFFFFFF, *(byte*)REG2180);
		printf8 (0,168, "$213C(%X)/$213D(%X) Scanline location (H/V) (dbl)", *REG213C, *REG213D);
		printf8 (0,176, "$4200(%X) Enable b7:NMI b5:V.Counter b4:H.Counter b0:joypad", *REG4200);
		printf8 (0,184, "$4207(%X)/$4209(%X) When to make IRQ int (H/V count)", *REG4207, *REG4209);
		printf8 (0,192, "$420B(%X)/$420C(%X) DMA/HDMA Transfer Initiate", *REG420B, *REG420C);
		printf8 (0,200, "$43?0(%X/%X/%X/%X / %X/%X/%X/%X) DMA Control", *REG4300(0), *REG4300(1), *REG4300(2), *REG4300(3), *REG4300(4), *REG4300(5), *REG4300(6), *REG4300(7));
		printf8 (0,208, "b7:PPU>CPU b6:HDMA ind b4:Dec b3:Fixed b0-2:Mode");
		printf8 (0,216, "$2101(%X) $2102(%X) OAM Size b0-2:base b5-7:size / Addr", *REG2101, *REG2102);
		printf8 (0,232, "OAM byte0:X 1:Y 2:Char 3:b0=CharMSB b1-3=Pal b4-5=Pri b6-7:flip");
	} else if (which == 8) {
		clearscreen ();
		for (y = 0; y < 56; y++) {
			for (x = 0; x < 16; x++) {
				if (vramblitoffset + (vramblitcdepth+1)*16*(y*16+x) >= 0x10000) {
					for (y2 = 0; y2 < 8; y2++)
						for (x2 = 0; x2 < 8; x2++)
							if (y < 28)
								vs[(y*8+y2) * xres + x*8+x2] = 30;
							else
								vs[((y-28)*8+y2) * xres + x*8+x2 + 128] = 30;
					continue;
				}
				tile = getcachetile ((vramblitoffset>>4) + (vramblitcdepth+1)*(y*16+x), vramblitcdepth, 0);
				for (y2 = 0; y2 < 8; y2++) {
					if (y < 28) {
						for (x2 = 0; x2 < 8; x2++) {
							vs[(y*8+y2) * xres + x*8+x2] = tile->pixel[y2*8+x2];
						}
					} else {
						for (x2 = 0; x2 < 8; x2++) {
							vs[((y-28)*8+y2) * xres + x*8+x2 + 128] = tile->pixel[y2*8+x2];
						}
					}
				}
			}
		}
		for (y = 0; y < 244; y++) {
			vs[y*xres+128] = messagecolor;
		}
		setcolor (46);
		printf8 (5, 230, "VRAM Blit: cdepth %d offs %d", vramblitcdepth, vramblitoffset);
		printf8 (5, 5, "$%X", vramblitoffset);
		printf8 (133, 5, "$%X", vramblitoffset + (vramblitcdepth+1)*16*28*16);
	}
}

