/*  CPU.HPP - 65c816 emulation stuff
*/

#define AUTOBACKFILESIZE 207000
#define AUTOBACKSPCSIZE 65600
extern byte *autobacktrackmem[10];
extern int autobackpos[9], autobacksize[9];
extern int autoback_countdown, autoback_restorecount;
size_t vfwrite(void *buf, size_t elsize, size_t nelem, FILE *fp);
size_t vfread (void *buf, size_t elsize, size_t nelem, FILE *fp);
FILE *vfopen (const char *filename, const char *mode);
void alloc_autobacktrack(int num);
void dealloc_autobacktrack();
void restorebacktrack ();
void savebacktrack ();
int vfeof (FILE *fp);
void vfclose (FILE *fp);
void dohdmaline ();
void savetallies ();

extern byte ram [0x20000], sram [0x10000], expram [0x2000];
extern byte vram [0x10000];
extern byte cgram [512], oam [544];

extern "C" {
	extern byte __cdecl registers [0x4000];
	extern int __cdecl frame_cycles, screen_scanline, total_frames, cycles_per_scan;
	extern int scan_cycles;
	extern byte __cdecl *startaddr [0x800];
	/*  For quick address aaccess.  To access a 65c816 address bb.aaaa, stored in variable addr:
		startaddr [addr >> 13] + (addr & 0x1FFF)
		This array is filled by getstartaddresses ().
	*/
	extern long __cdecl debug_instr;
	extern boolean returntogui;
	extern byte *rom;
}

extern boolean debugmode;

extern boolean showfps;

extern long framestarttime;
extern int snesframe;

extern int instantsaveslot;

struct headerdata {
	boolean hirom; // Apparent HiROM
	word hadd; // Set to 512
	char name [22];
	byte romsize; // Megabits
	byte sramsize; // Kilobits
	boolean fastrom;
	byte country, license, version;
	word resetvector, nmivector;
};

extern struct headerdata curheader; // Header of ROM that is currently loaded

#define SNESMEM(addr) (startaddr [(addr) >> 13] + ((addr) & 0x1FFF))

//byte *SNESMEM(int addr);

//#pragma aux SNESMEM = \
"mov edi,eax"\
"shr eax,11"\
"and al,0xFC"\
"and edi,01FFFh"\
"mov eax,[startaddr+eax]"\
"add edi,eax"\
parm [eax] modify [eax edi] value [edi];

//extern dword PC;
//extern word Y;
//extern word X;
//extern byte DBR;
//extern word D;
//extern word A;
//extern byte P;
//extern dword S;  
//extern boolean E;

extern struct opcodeinfo {
	char *format; // The first three bytes MUST BE THE OPCODE.
	// The description will be looked up according to the first three bytes.
	// S - stack relative
	// A - implicitly uses accumulator (no argument)
	// X and Y - Implied arguments specifying index registers X or Y.
	// %# - number.  This will be assumed to be a word if M=0, otherwise a byte.
	// %#i - number, but size is dependant on the X (Index) flag.
	// %#b - byte number.  Used with CLP and SEP.
	//      Numbers are automatically prefixed with "#".
	// %b - Argument is a 8-bit byte.
	// %w - Argument is a 16-bit word.
	// %L - Argument is a long address (3 bytes.)
	// @%b and @%w - Program counter reletive; signed byte or word.
	byte bytes, cycles;
} opcodelist[];

char *getaccessoryname (char *fn, char *dir, char *ext);
void savesram ();
void loadsram ();
boolean instantsave (int slot, boolean autobacktrack = false);
boolean instantload (int slot, boolean autobacktrack = false);

void resetsystem (boolean warmreset);
void getstartaddresses (boolean hirom);
void runsnes ();
void unassemble (dword opdata, char *s, byte P);
char *getdescr (byte opcode);

void startvblank ();
void startcpuframe ();
void startscanline ();
boolean ishirom ();
boolean ispal ();

extern "C" {
	byte __cdecl near *trapregread (byte *ptr, boolean wordread);
	byte __cdecl near *trapregwrite (byte *ptr, boolean wordwrite);
}

void togglecheat (int n);

/*  Apparent VRAM organization notes:
	2107 is the BG1 VRAM location register.  Bits 2-7 in this register specify
	the BYTE address of where the TILE MAP starts.  That is,
		LDA.b #4
		STA $2107  ; would mean that the tile map for BG1 starts at vram[2048]
	In other words, shift the value left by 11.
	Bits 0-1 are the "SC Size", meaning the dimensions of the tile map.
	Sizes are 0:32x32 1:32x64 2:64x32 3:64x64
	Each element in the tile map is two bytes and formatted like so:
		YXpc ccNN NNNN NNNN   Y = mirror y     X = mirror x   p=priority bit
		fedc ba98 7654 3210   ccc = palette nr (8 palettes)   NN.. = character number
	The location where the actual tiles start (for all four BGs) is selected
	with the registers $210B and $210C.  (Four bits per BG.)  This number
	is shifted left by 12 to get the final address.  At this location is an
	array of tiles, and the size of the elements of this array is calculated
	by: tile width * tile height * bit planes / 8 = bytes per tile
	eg in mode 0 it would be either 8*8*2/8=16, or 16x16x2/8=64.
	the palette is selected by bits 10-12 in the tile map; where a given
	palette number starts is given by: palette_number << bitplanes.
	
	Sprites: information about the sprites is stored in OAM.  Each sprite has
	4 bytes plus 2 bits of information about it in OAM.  This information is:
	Byte 0 xxxxxxxx  x: X-location.      c: Playfield priority.
	Byte 1 yyyyyyyy  y: Y-location.      d: Playfield priority.
	Byte 2 abcdeeeC  a: Vertical flip.   e: Pallete #.
	Byte 3 CCCCCCCC  b: Horizontal flip. C: Tile where the sprite data starts.
	Note: sprite palette 0 is the eighth palette (color 128?)
	The C bits index into another array of tiles.  This array of tiles starts
	at the location selected by bits 0-2 of $2101.  The value here is shifted
	left 14 (multiplied by 16384) to get the location in VRAM.  Add this to
	C multiplied by 32 to get the starting location of the sprite's bitmap.
	All tiles are made up of 8x8, 16 color (4 bit plane) sprites.  Sprites
	that are larger than 8x8 are made up of many of these 8x8 tiles; these
	tiles are organised left to right, and then top to bottom.
*/

/*REGISTERS*/

/*OAM (objects) REGISTERS**************************************************/
#define REG2101 ((byte*)(registers + 0x0101))
	// OAM Size register
	//Sprite sizes: b5-7: 000 =  8x8  or 16x16. 001 =  8x8  or 32x32.
	//                    010 =  8x8  or 64x64. 011 = 16x16 or 32x32.
	//                    100 = 16x16 or 64x64. 101 = 32x32 or 64x64.
	//b3-4: Name selection (upper 4k word addr).
	//b0-2: Base selection (8k word seg. addr).
#define REG2102 ((word*)(registers + 0x0102))
	// OAM address register b7 of loword is OAM priority rotation b0, is OAM address MSBit
	// hi word is OAM address LSByte
#define REG2104 ((byte*)(registers + 0x0104))
	// OAM Data write register
#define REG2138 ((byte*)(registers + 0x0138)) /*read-only*/
	// OAM Data read register

/*VIDEO REGISTERS/VRAM UPLOAD**********************************************/
#define REG2100 ((byte*)(registers + 0x0100))
	// Screen display register b7: 0=screen on 1=screen off
	// b0-3: brightness, 0xF = full
#define REG2105 ((byte*)(registers + 0x0105))
	// Screen mode register
#define REG2106 ((byte*)(registers + 0x0106))
	// Screen pixellation (Mosaic) b0-3 which BG to affect b4-7 pixel size
#define REG2107(plane) ((byte*)(registers + 0x0107 + plane))
	// VRAM Location register (BGxSC) (Tile maps)
	// $2107 to $210A b2-7:base address(*1K) b0-1:SC size
	// SC Sizes are 32x32, 32x64, 64x32 or 64x64. (In that order??)
#define REG210B ((word*)(registers + 0x010B))
	// Tile VRAM address (*8K, 4 bits for each BG)
#define REG210D(index) ((byte*)(registers + 0x010D + index))
	// Scroll registers & associated values--horizontal,vertical BG1; h,v BG2 etc
	// 8 double-write registers: first byte is low 8 bits, last byte b0-2:high 3 bits
	// (set to 0 except in mode 7) b3-7:mode 7 only
#define REG2115 ((byte*)(registers + 0x0115))
	// Video port control (VRAM upload settings control) b7:1=inc at 2119
#define REG2116 ((word*)(registers + 0x0116))
	// VRAM address register - write word
#define REG2118 ((word*)(registers + 0x0118))
	// VRAM write data 
#define REG2139 ((word*)(registers + 0x0139)) /*read-only*/
	// VRAM read data

/*MODE 7 REGISTERS*********************************************************/
#define REG211A ((byte*)(registers + 0x011A))
	// Mode 7 settings register
#define REG211B ((byte*)(registers + 0x011B))
#define REG211C ((byte*)(registers + 0x011C))
	// COS/SIN rotate angle/X expansion, $211B=16bit mult, $211C=8bit mult
#define REG211D ((byte*)(registers + 0x011D))
#define REG211E ((byte*)(registers + 0x011E))
	// SIN/COS rotate angle/Y expansion
#define REG211F ((byte*)(registers + 0x011F))
#define REG2120 ((byte*)(registers + 0x0120))
	// Center position X/Y (word) and associated values
#define REG2134 ((dword*)(registers + 0x0134)) /*read-only*/
	// 24-bit register--Mode 7 multiplication result

/*CGRAM (Palette) REGISTERS************************************************/
#define REG2121 ((byte*)(registers + 0x0121))
	// Palette (color #) selection register
#define REG2122 ((byte*)(registers + 0x0122))
	// Palette data write register
#define REG213B ((byte*)(registers + 0x013B)) /*read-only*/
	// Palette data read register

/*WINDOW REGISTERS*********************************************************/
#define REG2123 ((byte*)(registers + 0x0123))
	// Window mask settings register (BG2,BG1)
#define REG2124 ((byte*)(registers + 0x0124))
	// Window mask settings register (BG4,BG3)
#define REG2125 ((byte*)(registers + 0x0125))
	// Window mask settings register (Color,OBJ)
#define REG2126 ((byte*)(registers + 0x0126))
	// Window 1 left position
#define REG2127 ((byte*)(registers + 0x0127))
	// Window 1 right position
#define REG2128 ((byte*)(registers + 0x0128))
	// Window 2 left position
#define REG2129 ((byte*)(registers + 0x0129))
	// Window 2 right position
#define REG212A ((byte*)(registers + 0x012A))
	// Mask logic settings for BGs
#define REG212B ((byte*)(registers + 0x012B))
	// Mask logic settings for OBJs and Color windows

/*MORE VIDEO STUFF/COLOR ADD&SUB*******************************************/
#define REG212C ((byte*)(registers + 0x012C))
	// Main screen designation b0-4 Enable BGs/OAMs (b0=BG0, b4=sprites)
#define REG212D ((byte*)(registers + 0x012D))
	// Sub-screen designation b0-4 Enable BGs/OAMs (b0=BG0, b4=sprites)
#define REG212E ((byte*)(registers + 0x012E))
	// Window mask main screen designation
	// I'm guessing this must have a bit set to enable windows for that BG?
#define REG212F ((byte*)(registers + 0x012F))
	// Window mask sub screen designation
#define REG2130 ((byte*)(registers + 0x0130))
	// Fixed color add/subtraction register
#define REG2131 ((byte*)(registers + 0x0131))
	// in struct: byte reg2131setting [6];
	// Addition or subtraction select, etc and associated values
#define REG2132 ((byte*)(registers + 0x0132))
	// Fixed color add/sub data
#define REG2133 ((byte*)(registers + 0x0133))
inline int scans_before_vbl () {
	return (*REG2133 & 0x8) ? 239 : 224;
}
	// Screen mode b0:interlace b1:1=repeat OBJ dots(full height)
	// b2:0=224/1=239 lines b3:pseudo 512line mode b6:extbg(screen expand) b7:superimpose
#define REG2134 ((dword*)(registers + 0x0134)) /*read-only 24-bit*/
#define REG213E ((byte*)(registers + 0x013E)) /*read-only*/
	// PPU status flag/version number b7:time over b6:range over b5:master/slave b3-0:version
#define REG213F ((byte*)(registers + 0x013F)) /*read-only*/
	// PPU status flag/version number b7:field# in int.mode b6:ext.signal b4:1=PAL b0-3:version

/*MISC. REGISTERS**********************************************************/
#define REG2140 ((dword*)(registers + 0x0140)) /*read/write trapped*/
	// SPC700 communication ports
#define REG2180 ((byte*)(registers + 0x0180))
	// WRAM data - auto-increment
#define REG2181 ((dword*)(registers + 0x0181))
	// 24-bit WRAM address

/*COUNTER/TIMER REGISTERS**************************************************/
#define REG2137 ((byte*)(registers + 0x0137)) /*read-only*/
	// Sotware latch for horizontal/vertical counter, sets hcounter and vcounter.
#define REG213C ((byte*)(registers + 0x013C)) /*read-only? set by 2137*/
	//in struct boolean hhipointer;
	//word hcounter;
	// Horizontal scanline location and associated values
#define REG213D ((byte*)(registers + 0x013D)) /*read-only? set by 2137*/
	//boolean vhipointer;
	//word vcounter;
	// Vertical scanline location and associated values
#define REG4200 ((byte*)(registers + 0x2200)) /*read/write??*/
	// Counter enable b7:NMI interrupt b5:vertical IRQ b4:Horizontal IRQ b0:auto-read enable
	// if joypad is automatically read you can get the value from $4218 etc.
#define REG4207 ((word*)(registers + 0x2207))
	// Horizontal IRQ apply time (0-339;256-339 is overscan)
#define REG4209 ((word*)(registers + 0x2209))
	// Vertical IRQ apply time (o-261;224/239-261 is vblank)
#define REG4210 ((byte*)(registers + 0x2210))
	// This register gets read by the NMI handler to 'clear' the interrupt (i.e. turn off the
	// NMI request in the circuitry.) returns b7=1 when read, then b7=0 when it's read again. (?)
	// b0-3=version number
#define REG4211 ((byte*)(registers + 0x2211))
	// This register gets read by the IRQ handler to 'clear' the interrupt (i.e. turn off the
	// IRQ request in the circuitry.) returns b7=1 when read, then b7=0 when it's read again. (?)
#define REG4212 ((byte*)(registers + 0x2212))
	// Status register. b7=whether in VBlank b6=Whether in HBlank D0=joypad is ready to read?

/*MISC AND MATH REGISTERS**************************************************/
#define REG4201 ((byte*)(registers + 0x2201))
	// Programmable I/O Port output
#define REG4202 ((byte*)(registers + 0x2202))
	// Multiplicand 'A'--8-bit (A is set first)
#define REG4203 ((byte*)(registers + 0x2203))
	// Multiplier 'B'--8-bit (B is set last)
#define REG4204 ((word*)(registers + 0x2204))
	// Dividend 'C'--16-bit
#define REG4206 ((byte*)(registers + 0x2206))
	// Divisor-'B'--8-bit
#define REG420D ((byte*)(registers + 0x220D))
	// Cycle speed b0:FastROM
#define REG4213 ((byte*)(registers + 0x2213))
	// Programmable I/O Port (input)
#define REG4214 ((word*)(registers + 0x2214))
	// 16-bit quotient
#define REG4216 ((word*)(registers + 0x2216))
	// 16-bit multiplication result, or remainder

/*JOYPAD READING REGISTERS*************************************************/
#define REG4218(joypad) ((word*)(registers + 0x2218 + joypad*2))
	// Joypad status registers: low byte: b7=A b6=X b5=L button b4=R button;
	// high byte: b7=B b6=Y b5=Select b4=Start b3=Up b2=Down b1=Left b0=Right
	// 4218=Joypad 1 421A=Joypad 2 421C=Joypad 3 421E=Joypad 4
#define REG4016 ((byte*)(registers + 0x2016))
	// Old-style joypad. When read: b0:Data for controller 1 b1:Data for controller 2
	// When written:??? (Trigger strobe?) (17th bit returns whether joypad is connected)
#define REG4017 ((byte*)(registers + 0x2017))
	// b0:Data for controller 3 b1:data for controller 4??
#define REG420B ((byte*)(registers + 0x220B))
	// Set bits to initiate DMA transfer
#define REG420C ((byte*)(registers + 0x220C))
	// Same as $420B, but to enable automatic HDMA transfers
#define REG4300(chan) ((byte*)(registers + 0x2300 + (chan << 4)))
	// DMA Control Register b7:(DMA)0=CPU->PPU b6:(HDMA)0=Absolute 1=Indirect addressing
	// b4:0=Inc/Dec Addresses 1=Fixed Address b3:0=Increment 1=Decrement b0-2:transfer mode

#define REG4301(chan) ((byte*)(registers + 0x2301 + (chan << 4)))
	// $21xx address
#define REG4302(chan) ((dword*)(registers + 0x2302 + (chan << 4)))
	// 24-bit CPU address
#define REG4305(chan) ((word*)(registers + 0x2305 + (chan << 4)))
	// Bytes to transfer (DMA) or some kind of 24-bit CPU address (HDMA)
#define REG4307(chan) ((byte*)(registers + 0x2307 + (chan << 4)))
#define REG4308(chan) ((word*)(registers + 0x2308 + (chan << 4)))
#define REG430A(chan) ((byte*)(registers + 0x230A + (chan << 4)))
	// don't know what 4307+ are for
#define REG ((byte*)(registers + 0x))

#pragma pack (push,4)
extern "C" {
	extern struct ppuinternalregisters {  /* DO NOT CHANGE ORDER OF DATA! */
		boolean end_wai, within_wai; // flags to control WAI command behavior
		byte reg4210, reg4211;
		word scrollreg[8];
		byte reg2131setting [6]; // 0-3 BGs, 4 OBJs, 5 back area
		word hcounter, vcounter;
		byte reg2101;
		int vramincrate;
		signed short reg211B; // Mode 7 multiply 16bit
		signed char reg211C;  // Mode 7 multiply 8bit
		byte hdma_inprogress;
		short m7var[6];
		int oampointer, colorpointer;
		int vrampointer;
		int oldjoybit[2]; // Bit # in old joypad regs ($4016/$4017), incrememted after read
		boolean dummyread2139; // true if dummy read required (before "real" read on SNES VRAM)
	} state;
}
#pragma pack(pop)
